# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Clone collection repos
"""

from __future__ import annotations

import asyncio
import dataclasses
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import aiofiles.os
import aiofiles.ospath
from antsibull_core.logging import log
from antsibull_core.subprocess_util import async_log_run
from antsibull_core.yaml import load_yaml_file, store_yaml_file

from antsibull.tagging import CollectionTagData
from antsibull.types import CollectionName
from antsibull.utils.paths import copytree_and_symlinks

from .exceptions import CloneError
from .verify import FileError, FileErrorOutput

if TYPE_CHECKING:
    from _typeshed import StrPath

mlog = log.fields(mod=__name__)


async def clone_collection(
    collection: CollectionName,
    tag_data: CollectionTagData,
    download_dir: Path,
    *,
    depth: int | None = 1,
) -> Path:
    """
    Clone a collection's git repository

    Args:
        tag_data:
            A tag dictionary. See `antsibull.tagging.get_collections_tags()`.
    Returns:
        The path to the collection checkout
    """
    flog = mlog.fields(func="clone_collection")
    if not all((tag_data["repository"], tag_data["tag"])):
        raise CloneError(
            "'tag_data' does not specify a 'repository' and 'tag'", collection
        )
    dest = download_dir / collection.namespace / collection.name
    await asyncio.to_thread(dest.parent.mkdir, exist_ok=True)
    args: list[StrPath] = [
        "git",
        "clone",
        f"--branch={tag_data['tag']}",
        cast(str, tag_data["repository"]),
        dest,
    ]
    if depth is not None:
        args.insert(2, f"--depth={depth}")
    await async_log_run(args, logger=flog)
    return dest


@dataclasses.dataclass
class NormalizedCheckout:
    """
    Dataclass that represents a collection checkout

    Attributes:
        collection:
            The collection namespace/name
        clone_directory:
            Path to the collection's git checkout
        normalized_data:
            A normalized version of the galaxy.yml
        data_changed:
            Whether the metadata was modified during normalization
        collection_directory:
            Path to the collection directory with galaxy.yml.
            If not passed and clone_directory ends with
            `ansible_collections/NAMESPACE/NAME`, this is set to
            `clone_directory`.
    """

    collection: CollectionName
    tag_data: CollectionTagData
    clone_directory: Path
    normalized_data: dict[str, Any]
    data_changed: bool
    collection_directory: Path | None
    errors: list[FileErrorOutput] = dataclasses.field(default_factory=list)

    def __post_init__(self) -> None:
        self.collection_directory = self._guess_collection_directory()

    @property
    def collection_subdir(self) -> Path:
        return Path(self.tag_data.get("collection_directory", "./"))

    def _guess_collection_directory(self) -> Path | None:
        if not self.collection_directory:
            collection_subdir = self.clone_directory / self.collection_subdir
            if _matches_collection(collection_subdir, self.collection):
                return collection_subdir
        else:
            _matches_collection_err(self.collection_directory, self.collection)
        return self.collection_directory

    async def add_to_collection_tree(
        self, tree_path: Path, allow_symlink: bool = True
    ) -> Path:
        dest = Path(tree_path, *self.collection.parts)
        _matches_collection_err(dest, self.collection)
        await asyncio.to_thread(dest.parent.mkdir, exist_ok=True)
        if self.collection_directory and allow_symlink:
            await aiofiles.os.symlink(self.collection_directory.absolute(), dest)
        else:
            await asyncio.to_thread(
                copytree_and_symlinks,
                self.clone_directory / self.collection_subdir,
                dest,
            )
        return dest


def _matches_collection(directory: Path, collection: CollectionName) -> bool:
    return directory.parts[-3:] == ("ansible_collections", *collection.parts)


def _matches_collection_err(directory: Path, collection: CollectionName) -> None:
    if not _matches_collection(directory, collection):
        msg = f"{directory!r} does not end in ansible_collections/NAMESPACE/NAME"
        raise ValueError(msg)


async def normalize_clone(
    collection: CollectionName,
    checkout_dir: Path,
    tag_data: CollectionTagData,
) -> NormalizedCheckout:
    """
    Normalize a collection checkout. Ensure that the collection namespace,
    name, and version in galaxy.yml are correct.

    Args:
        checkout_dir:
            Path to the collection git checkout
        expected_collection:
            Expected NAMESPACE.NAME of the collection
        tag_data:
            A tag dictionary. See `antsibull.tagging.get_collections_tags()`.
    """
    flog = mlog.fields(func="normalize_clone", collection=collection)
    errors: list[FileErrorOutput] = []

    collection_directory = checkout_dir / tag_data.get("collection_directory", "./")
    galaxy_path = collection_directory / "galaxy.yml"
    if not await aiofiles.ospath.isfile(galaxy_path):
        raise CloneError(f"{galaxy_path} does not exist!", collection)
    galaxy_data: dict[str, Any] = load_yaml_file(galaxy_path)
    new_data = galaxy_data.copy()

    gotten_collection = f"{galaxy_data.get('namespace')}.{galaxy_data.get('name')}"
    if gotten_collection != collection:
        new_data["namespace"], new_data["name"] = collection.parts
        msg = (
            "Divergent collection names found."
            f" expected collection is {collection!r}"
            f" and galaxy metadata says {gotten_collection!r}"
        )
        flog.warning(msg)
        errors.append(
            {
                "file": "galaxy.yml",
                "error": FileError.WRONG_GALAXY_YML_NAME,
                "message": f"`{gotten_collection}` != `{collection}`",
            }
        )

    # Some collections set version to nil and change it dynamically...
    # Others don't set version at all, hence the .get()
    if galaxy_data.get("version") is None:
        new_data["version"] = tag_data["version"]
    elif new_data["version"] != tag_data["version"]:
        new_data["version"] = tag_data["version"]
        msg = (
            "Divergent collection versions found."
            f" expected version is {tag_data['version']!r}"
            f" and galaxy metadata says {galaxy_data['version']!r}"
        )
        flog.warning(msg)

    if changed := galaxy_data != new_data:
        store_yaml_file(galaxy_path, new_data)
    return NormalizedCheckout(
        collection, tag_data, checkout_dir, new_data, changed, None, errors
    )


__all__ = ("clone_collection", "normalize_clone", "NormalizedCheckout")



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2023
"""Handle version specific things."""

from __future__ import annotations

import asyncio
import os
import sys
import typing as t
from collections.abc import Mapping, Sequence

import aiohttp
import asyncio_pool  # type: ignore[import]
from antsibull_core import app_context
from antsibull_core.ansible_core import AnsibleCorePyPiClient
from antsibull_core.dependency_files import parse_pieces_file
from antsibull_core.galaxy import GalaxyClient, GalaxyContext
from packaging.version import Version as PypiVer
from semantic_version import SimpleSpec as SemVerSpec
from semantic_version import Version as SemVer


class _FeatureFreezeVersion:
    """
    Helper for making semantic version range specification valid for feature freeze.
    """

    def __init__(self, spec: str, collection_name: str):
        self.potential_clauses: list = []
        self.spec = spec
        self.collection_name = collection_name
        self.upper_operator: str | None = None
        self.upper_version: SemVer | None = None
        self.min_version: SemVer | None = None
        self.pinned = False

        spec_obj = SemVerSpec(spec)

        # If there is a single clause, it's available as spec_obj.clause;
        # multiple ones are available as spec_obj.clause.clauses.
        try:
            clauses = spec_obj.clause.clauses
        except AttributeError:
            clauses = [spec_obj.clause]

        self.clauses = clauses
        for clause in clauses:
            self._process_clause(clause)

    def _process_clause(self, clause) -> None:
        """
        Process one clause of the version range specification.
        """
        if clause.operator in ("<", "<="):
            if self.upper_operator is not None:
                raise ValueError(
                    f"Multiple upper version limits specified for {self.collection_name}:"
                    f" {self.spec}"
                )
            self.upper_operator = clause.operator
            self.upper_version = clause.target
            # Omit the upper bound as we're replacing it
            return

        if clause.operator == ">=":
            # Save the lower bound so we can write out a new compatible version
            if self.min_version is not None:
                raise ValueError(
                    f"Multiple minimum versions specified for {self.collection_name}: {self.spec}"
                )
            self.min_version = clause.target

        if clause.operator == ">":
            raise ValueError(
                f"Strict lower bound specified for {self.collection_name}: {self.spec}"
            )

        if clause.operator == "==":
            self.pinned = True

        self.potential_clauses.append(clause)

    def compute_new_spec(self) -> str:
        """
        Compute a new version range specification that only allows newer patch releases that also
        match the original range specification.
        """
        if self.pinned:
            if len(self.clauses) > 1:
                raise ValueError(
                    f"Pin combined with other clauses for {self.collection_name}: {self.spec}"
                )
            return self.spec

        upper_operator = self.upper_operator
        upper_version = self.upper_version
        if upper_operator is None or upper_version is None:
            raise ValueError(
                f"No upper version limit specified for {self.collection_name}: {self.spec}"
            )

        min_version = self.min_version
        if min_version is None:
            raise ValueError(
                f"No minimum version specified for {self.collection_name}: {self.spec}"
            )

        if min_version.next_minor() <= upper_version:
            upper_operator = "<"
            upper_version = min_version.next_minor()

        new_clauses = sorted(
            str(clause)
            for clause in self.potential_clauses
            if clause.target < upper_version
        )
        new_clauses.append(f"{upper_operator}{upper_version}")
        return ",".join(new_clauses)


def feature_freeze_version(spec: str, collection_name: str) -> str:
    """
    Make semantic version range specification valid for feature freeze.
    """
    return _FeatureFreezeVersion(spec, collection_name).compute_new_spec()


def load_constraints_if_exists(
    filename: str | os.PathLike[str],
) -> dict[str, SemVerSpec]:
    """
    Load a constraints file, if it exists.
    """
    result: dict[str, SemVerSpec] = {}
    if not os.path.exists(filename):
        return result
    for line in parse_pieces_file(filename):
        record = [entry.strip() for entry in line.split(":", 1)]
        if len(record) < 2:
            raise ValueError(
                f'While parsing {filename}: record "{line}" is not of the form "collection: spec"'
            )
        collection = record[0]
        try:
            constraint = SemVerSpec(record[1])
        except ValueError as exc:
            raise ValueError(
                f"While parsing {filename}: cannot parse constraint"
                f' "{record[1]}" for collection {collection}: {exc}'
            ) from exc
        result[collection] = constraint
    return result


def _display_exception(loop, context):  # pylint:disable=unused-argument
    print(context.get("exception"), file=sys.stderr)


async def get_version_info(
    collections: Sequence[str],
    pypi_server_url: str | None = None,
    *,
    galaxy_context: GalaxyContext,
) -> tuple[dict[str, t.Any], dict[str, list[str]]]:
    """
    Return the versions of all the collections and ansible-core
    """
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(_display_exception)

    requestors = {}
    lib_ctx = app_context.lib_ctx.get()
    async with (
        aiohttp.ClientSession() as aio_session,
        asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool,
    ):
        pypi_client = AnsibleCorePyPiClient(
            aio_session, pypi_server_url=pypi_server_url
        )
        requestors["_ansible_core"] = await pool.spawn(pypi_client.get_release_info())

        galaxy_client = GalaxyClient(aio_session, context=galaxy_context)
        for collection in collections:
            requestors[collection] = await pool.spawn(
                galaxy_client.get_versions(collection)
            )

        collections_to_versions = {}
        responses = await asyncio.gather(*requestors.values())

    ansible_core_release_infos: dict[str, t.Any] | None = None
    for idx, collection_name in enumerate(requestors):
        if collection_name == "_ansible_core":
            ansible_core_release_infos = responses[idx]
        else:
            collections_to_versions[collection_name] = responses[idx]

    if ansible_core_release_infos is None:
        raise RuntimeError("Internal error")

    return ansible_core_release_infos, collections_to_versions


def get_latest_ansible_core_version(
    ansible_core_versions: Sequence[str],
    ansible_core_version: PypiVer,
    pre: bool = False,
) -> PypiVer | None:
    """
    Retrieve the latest ansible-core bugfix release's version for the given ansible-core version.

    :arg ansible_core_versions: A list of ansible-core versions.
    :arg ansible_core_version: A ansible-core version.
    """
    versions = [PypiVer(v) for v in ansible_core_versions]
    next_version = PypiVer(
        f"{ansible_core_version.major}.{ansible_core_version.minor + 1}a"
    )
    newer_versions = [
        version
        for version in versions
        if ansible_core_version <= version < next_version
        and (pre or not version.is_prerelease)
    ]
    return max(newer_versions) if newer_versions else None


def get_latest_collection_version(
    versions: Sequence[str],
    collection: str,
    version_spec: str | None = None,
    pre: bool = False,
    prefer_pre: bool = False,
    constraint: SemVerSpec | None = None,
) -> SemVer:
    """
    Get the latest version of a collection that matches a specification.

    :arg versions: Sequence of collection versions
    :arg collection: Namespace.collection identifying a collection.
    :arg version_spec: Optional string specifying the allowable versions.
    :kwarg pre: If ``True``, allow prereleases (versions which have the form X.Y.Z.SOMETHING).
        This is **not** for excluding 0.Y.Z versions.  Non-pre-releases are still
        preferred over pre-releases, except if ``prefer_pre=True`` (for instance, with
        ``version_spec='>2.0.0-a1,<3.0.0'`` and ``pre=True``, if the available versions
        are 2.0.0-a1 and 2.0.0-a2, then 2.0.0-a2 will be returned.  If the available
        versions are 2.0.0 and 2.1.0-b2, 2.0.0 will be returned since non-pre-releases
        are preferred.) The default is ``False``.
    :kwarg prefer_pre: If ``True``, prefer newer pre-releases over stable releases. Is only
        used when ``pre=True``.
    :kwarg constraint: If provided, only consider versions that match this specification.
    :returns: :obj:`semantic_version.Version` of the latest collection version that satisfied
        the specification.

    .. seealso:: For the format of the version_spec, see the documentation
        of :obj:`semantic_version.SimpleSpec`
    """
    sem_versions = [SemVer(v) for v in versions]
    sem_versions.sort(reverse=True)

    if version_spec:
        spec = SemVerSpec(version_spec)
        sem_versions = [v for v in sem_versions if v in spec]

    if constraint:
        # Ignore all versions that do not match the constraints
        sem_versions = [v for v in sem_versions if v in constraint]

    prereleases = []
    for version in sem_versions:
        # If this is a pre-release, first check if there's a non-pre-release that
        # will satisfy the version_spec.
        if version.prerelease:
            # If we prefer prereleases, take this one.
            if pre and prefer_pre:
                return version
            prereleases.append(version)
            continue
        return version

    # We did not find a stable version that satisies the version_spec.  If we
    # allow prereleases, return the latest of those here.
    if pre and prereleases:
        return prereleases[0]

    # No matching versions were found
    constraint_clause = "" if constraint is None else f" (with constraint {constraint})"
    raise ValueError(
        f"{version_spec}{constraint_clause} did not match with any version of {collection}."
    )


def find_latest_compatible(
    ansible_core_version: PypiVer,  # pylint: disable=unused-argument
    collections_to_versions: Mapping[str, Sequence[str]],
    pre: bool = False,
    prefer_pre: bool = False,
    version_specs: Mapping[str, str] | None = None,
    constraints: Mapping[str, SemVerSpec] | None = None,
) -> dict[str, SemVer]:
    """
    Finds the latest compatible version of every collection from ``collections_to_versions``
    that matches the given ``vresion_specs`` and ``constraints``.

    ``pre`` and ``prefer_pre`` control whether pre-releases are acceptable (``pre=True``),
    and in case both matching pre-releases and releases are found, which ones to prefer
    (``prefer_pre=True`` prefers pre-releases over regular releases).
    """
    # Note: ansible-core compatibility is not currently implemented.  It will be a piece of
    # collection metadata that is present in the collection but may not be present in Galaxy.
    # We'll have to figure that out once the pieces are finalized

    version_specs = version_specs or {}
    constraints = constraints or {}

    reduced_versions = {}
    for dep, versions in collections_to_versions.items():
        reduced_versions[dep] = get_latest_collection_version(
            versions,
            dep,
            version_spec=version_specs.get(dep),
            pre=pre,
            prefer_pre=prefer_pre,
            constraint=constraints.get(dep),
        )

    return reduced_versions



# Copyright (c) Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import operator as py_operator

from packaging.version import parse


def packaging_version(version_a, version_b, op):
    a = parse(version_a)
    b = parse(version_b)
    method = getattr(py_operator, {
        '<': 'lt',
        '<=': 'le',
        '==': 'eq',
        '>=': 'ge',
        '>': 'gt',
    }[op])
    return method(a, b)


class TestModule:
    ''' Version jinja2 tests '''

    def tests(self):
        return {
            '_antsibull_packaging_version': packaging_version,
        }



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020
"""Build Ansible packages."""

from __future__ import annotations

import asyncio
import datetime
import os
import os.path
import shutil
import sys
import tempfile
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING

import aiohttp
import asyncio_pool  # type: ignore[import]
from antsibull_core import app_context
from antsibull_core.ansible_core import get_ansible_core
from antsibull_core.collections import install_together
from antsibull_core.dependency_files import BuildFile, DependencyFileData, DepsFile
from antsibull_core.galaxy import CollectionDownloader, GalaxyContext
from antsibull_core.logging import log
from antsibull_core.subprocess_util import async_log_run, log_run
from antsibull_core.yaml import store_yaml_file, store_yaml_stream
from jinja2 import Template
from packaging.version import Version as PypiVer
from semantic_version import Version as SemVer

from antsibull.constants import MINIMUM_ANSIBLE_VERSIONS
from antsibull.python_metadata import BuildMetaMaker, LegacyBuildMetaMaker

from . import __version__ as antsibull_version
from .build_changelog import ReleaseNotes
from .changelog import ChangelogData, get_changelog
from .dep_closure import check_collection_dependencies
from .tagging import get_collections_tags
from .utils.galaxy import create_galaxy_context
from .utils.get_pkg_data import get_antsibull_data
from .versions import (
    feature_freeze_version,
    find_latest_compatible,
    get_latest_ansible_core_version,
    get_version_info,
    load_constraints_if_exists,
)

if TYPE_CHECKING:
    from _typeshed import StrPath


mlog = log.fields(mod=__name__)

TAG_FILE_MESSAGE = """\
# This is a mapping of collections to their git repositories and the git tag
# that corresponds to the version included in this ansible release. A null
# 'tag' field means that a collection's release wasn't tagged.
"""


#
# Common code
#


async def download_collections(
    versions: Mapping[str, SemVer],
    download_dir: str,
    galaxy_context: GalaxyContext,
    collection_cache: str | None = None,
) -> dict[str, str]:
    requestors = {}
    async with aiohttp.ClientSession() as aio_session:
        lib_ctx = app_context.lib_ctx.get()
        async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
            downloader = CollectionDownloader(
                aio_session,
                download_dir,
                collection_cache=collection_cache,
                context=galaxy_context,
            )
            for collection_name, version in versions.items():
                requestors[collection_name] = await pool.spawn(
                    downloader.download(collection_name, version)
                )

            values = await asyncio.gather(*requestors.values())
    return dict(zip(versions.keys(), values))


async def _get_ansible_core_path(
    download_dir: StrPath, ansible_core_version: PypiVer | str
) -> Path:
    flog = mlog.fields(func="_get_ansible_core_path")
    async with aiohttp.ClientSession() as aio_session:
        ansible_core_tarball = Path(
            await get_ansible_core(
                aio_session, str(ansible_core_version), str(download_dir)
            )
        )
    await async_log_run(
        ["tar", "-C", download_dir, "-xf", ansible_core_tarball],
        logger=flog,
    )
    return Path(download_dir, ansible_core_tarball.with_suffix("").with_suffix("").name)


#
# Single sdist for ansible
#


def copy_boilerplate_files(package_dir: StrPath) -> None:
    gpl_license = get_antsibull_data("gplv3.txt")
    with open(os.path.join(package_dir, "COPYING"), "wb") as f:
        f.write(gpl_license)

    readme = get_antsibull_data("ansible-readme.rst")
    with open(os.path.join(package_dir, "README.rst"), "wb") as f:
        f.write(readme)

    pyproject_toml = get_antsibull_data("pyproject.toml")
    with open(os.path.join(package_dir, "pyproject.toml"), "wb") as f:
        f.write(pyproject_toml)


def write_manifest(
    package_dir: StrPath,
    release_notes: ReleaseNotes | None = None,
    debian: bool = False,
    tags_file: StrPath | None = None,
) -> None:
    manifest_file = os.path.join(package_dir, "MANIFEST.in")
    with open(manifest_file, "w", encoding="utf-8") as f:
        f.write("include COPYING\n")
        f.write("include README.rst\n")
        f.write("include build-ansible.sh\n")
        if release_notes:
            for changelog in release_notes.changelogs:
                f.write(f"include {changelog.filename}\n")
            f.write(f"include {release_notes.porting_guide.filename}\n")
        if debian:
            f.write("include debian/*\n")
        if tags_file:
            f.write("include tags.yaml\n")
        f.write("recursive-include ansible_collections/ **\n")


def write_release_py(ansible_version: PypiVer, ansible_collections_dir: str) -> None:
    release_filename = os.path.join(ansible_collections_dir, "ansible_release.py")

    release_tmpl = Template(get_antsibull_data("ansible-release_py.j2").decode("utf-8"))
    release_contents = release_tmpl.render(version=ansible_version)

    with open(release_filename, "w", encoding="utf-8") as f:
        f.write(release_contents)


def write_ansible_community_py(
    ansible_version: PypiVer, ansible_collections_dir: str
) -> None:
    release_filename = os.path.join(ansible_collections_dir, "ansible_community.py")

    release_tmpl = Template(
        get_antsibull_data("ansible-community.py.j2").decode("utf-8")
    )
    release_contents = release_tmpl.render(version=ansible_version)

    with open(release_filename, "w", encoding="utf-8") as f:
        f.write(release_contents + "\n")


def copy_tags_file(tags_file: StrPath | None, package_dir: StrPath) -> None:
    if tags_file:
        dest = os.path.join(package_dir, "tags.yaml")
        shutil.copy(tags_file, dest)


def write_python_build_files(
    package_dir: StrPath,
    release_notes: ReleaseNotes | None = None,
    debian: bool = False,
    tags_file: StrPath | None = None,
    stub_setup_py: bool = False,
) -> None:
    copy_boilerplate_files(package_dir)
    copy_tags_file(tags_file, package_dir)
    write_manifest(package_dir, release_notes, debian, tags_file)
    if stub_setup_py:
        Path(package_dir, "setup.py").write_bytes(
            get_antsibull_data("ansible-stub-setup.py")
        )


def write_debian_directory(
    ansible_version: PypiVer,
    ansible_core_version: PypiVer,  # pylint: disable=unused-argument
    package_dir: StrPath,
) -> None:
    debian_dir = os.path.join(package_dir, "debian")
    if not os.path.isdir(debian_dir):
        os.mkdir(debian_dir, mode=0o700)
    debian_files = ("changelog.j2", "control.j2", "copyright", "rules")
    for filename in debian_files:
        # Don't use os.path.join here, the get_data docs say it should be
        # slash-separated.
        src_pkgfile = "debian/" + filename
        data = get_antsibull_data(src_pkgfile).decode("utf-8")

        if filename.endswith(".j2"):
            filename = filename.replace(".j2", "")
            # If the file is a template, send it in vars it might need
            # and update 'data' to be the result.
            tmpl = Template(data)
            data = tmpl.render(
                version=str(ansible_version),
                date=datetime.datetime.utcnow().strftime("%a, %d %b %Y %T +0000"),
                ansible_core_package_name="ansible-core",
            )

        with open(os.path.join(debian_dir, filename), "w", encoding="utf-8") as f:
            f.write(data)


def write_galaxy_requirements(
    filename: str, included_versions: Mapping[str, str]
) -> None:
    galaxy_reqs = []
    for collection, version in sorted(included_versions.items()):
        galaxy_reqs.append(
            {
                "name": collection,
                "version": version,
                "source": "https://galaxy.ansible.com",
            }
        )

    store_yaml_file(
        filename,
        {
            "collections": galaxy_reqs,
        },
    )


def make_dist_with_wheels(ansible_dir: str, dest_dir: str) -> None:
    # TODO: build has an API, but it's quite unstable, so we use the cli for now
    log_run(
        [sys.executable, "-m", "build", "--outdir", dest_dir, ansible_dir],
        logger=mlog.fields(func="make_dist_with_wheels"),
        stderr_loglevel="warning",
    )


def write_build_script(
    ansible_version: PypiVer, ansible_core_version: PypiVer, package_dir: StrPath
) -> None:
    """Write a build-script that tells how to build this tarball."""
    build_ansible_filename = os.path.join(package_dir, "build-ansible.sh")

    build_ansible_tmpl = Template(
        get_antsibull_data("build-ansible.sh.j2").decode("utf-8")
    )
    build_ansible_contents = build_ansible_tmpl.render(
        version=ansible_version,
        ansible_core_version=ansible_core_version,
        antsibull_version=antsibull_version,
    )

    with open(build_ansible_filename, "w", encoding="utf-8") as f:
        f.write(build_ansible_contents)
    os.chmod(build_ansible_filename, mode=0o755)


def write_all_build_files(
    *,
    package_dir: StrPath,
    collections_dir: StrPath,
    ansible_version: PypiVer,
    dependency_data: DependencyFileData,
    ansible_core_version: PypiVer,
    python_requires: str | None = None,
    tags_path: StrPath | None = None,
    debian: bool,
    sdist_src_dir: StrPath | None = None,
    ansible_core_checkout: StrPath,
    release_notes: ReleaseNotes | None = None,
) -> None:
    use_build_meta_maker = (
        ansible_version >= MINIMUM_ANSIBLE_VERSIONS["BUILD_META_MAKER"]
    )
    meta_maker_class: type[BuildMetaMaker | LegacyBuildMetaMaker] = (
        BuildMetaMaker if use_build_meta_maker else LegacyBuildMetaMaker
    )
    build_meta_maker = meta_maker_class(
        package_dir=package_dir,
        collections_dir=collections_dir,
        ansible_version=ansible_version,
        dependency_data=dependency_data,
        ansible_core_version=ansible_core_version,
        ansible_core_checkout=ansible_core_checkout,
        python_requires=python_requires,
    )
    build_meta_maker.write()

    write_build_script(ansible_version, ansible_core_version, package_dir)
    write_python_build_files(
        package_dir, release_notes, debian, tags_path, use_build_meta_maker
    )
    if debian:
        write_debian_directory(ansible_version, ansible_core_version, package_dir)

    if sdist_src_dir:
        shutil.copytree(
            package_dir,
            sdist_src_dir,
            symlinks=True,
            ignore_dangling_symlinks=True,
        )


def build_single_command() -> int:
    # This is deprecated; in the future users will have to first run prepare, and then
    # rebuild-single.
    result = prepare_command()
    if result != 0:
        return result

    return rebuild_single_command()


def _is_alpha(version: PypiVer) -> bool:
    """Test whether the provided version is an alpha version."""
    pre = version.pre
    return version.is_prerelease and pre is not None and pre[0] == "a"


def _extract_python_requires(
    ansible_core_version: PypiVer, deps: dict[str, str]
) -> str:
    python_requires = deps.pop("_python", None)
    if python_requires is not None:
        return python_requires
    if ansible_core_version < PypiVer("2.14.0a"):
        # ansible-core 2.12 and 2.13 support Python 3.8+
        return ">=3.8"
    if ansible_core_version < PypiVer("2.15.0a"):
        # ansible-core 2.14 supports Python 3.9+
        return ">=3.9"
    raise ValueError(
        f"Python requirements for ansible-core {ansible_core_version} should be part of"
        " dependency information"
    )


def prepare_command() -> int:
    app_ctx = app_context.app_ctx.get()
    lib_ctx = app_context.lib_ctx.get()

    build_filename = os.path.join(
        app_ctx.extra["data_dir"], app_ctx.extra["build_file"]
    )
    build_file = BuildFile(build_filename)
    build_ansible_version, ansible_core_version, deps = build_file.parse()
    ansible_core_version_obj = PypiVer(ansible_core_version)
    python_requires = _extract_python_requires(ansible_core_version_obj, deps)

    constraints_filename = os.path.join(
        app_ctx.extra["data_dir"], app_ctx.extra["constraints_file"]
    )
    constraints = load_constraints_if_exists(constraints_filename)

    # If we're building a feature frozen release (betas and rcs) then we need to
    # change the upper version limit to not include new features.
    if app_ctx.extra["feature_frozen"]:
        old_deps, deps = deps, {}
        for collection_name, spec in old_deps.items():
            deps[collection_name] = feature_freeze_version(spec, collection_name)

    galaxy_context = asyncio.run(create_galaxy_context())
    ansible_core_release_infos, collections_to_versions = asyncio.run(
        get_version_info(
            list(deps),
            pypi_server_url=str(lib_ctx.pypi_url),
            galaxy_context=galaxy_context,
        )
    )

    new_ansible_core_version = get_latest_ansible_core_version(
        list(ansible_core_release_infos),
        ansible_core_version_obj,
        pre=_is_alpha(app_ctx.extra["ansible_version"]),
    )
    if new_ansible_core_version:
        ansible_core_version_obj = new_ansible_core_version

    included_versions = find_latest_compatible(
        ansible_core_version_obj,
        collections_to_versions,
        version_specs=deps,
        pre=True,
        prefer_pre=False,
        constraints=constraints,
    )

    if not str(app_ctx.extra["ansible_version"]).startswith(build_ansible_version):
        print(
            f"{build_filename} is for version {build_ansible_version} but we need"
            f' {app_ctx.extra["ansible_version"].major}'
            f'.{app_ctx.extra["ansible_version"].minor}'
        )
        return 1

    dependency_data = DependencyFileData(
        str(app_ctx.extra["ansible_version"]),
        str(ansible_core_version_obj),
        {collection: str(version) for collection, version in included_versions.items()},
    )

    # Get Ansible changelog, add new release
    ansible_changelog = ChangelogData.ansible(
        app_ctx.extra["data_dir"], app_ctx.extra["dest_data_dir"]
    )
    date = datetime.date.today()
    ansible_changelog.add_ansible_release(
        str(app_ctx.extra["ansible_version"]),
        date,
        f"Release Date: {date}"
        f"\n\n"
        f"`Porting Guide <https://docs.ansible.com/ansible/devel/porting_guides.html>`_",
    )
    ansible_changelog.changes.save()

    # Write dependency file
    deps_filename = os.path.join(
        app_ctx.extra["dest_data_dir"], app_ctx.extra["deps_file"]
    )
    deps_file = DepsFile(deps_filename)
    deps_file.write(
        dependency_data.ansible_version,
        dependency_data.ansible_core_version,
        dependency_data.deps,
        python_requires=python_requires,
    )

    # Write tags data
    if app_ctx.extra["tags_file"]:
        tag_data = asyncio.run(
            get_collections_tags(
                app_ctx.extra["dest_data_dir"], app_ctx.extra["deps_file"]
            )
        )
        tags_path = os.path.join(
            app_ctx.extra["dest_data_dir"], app_ctx.extra["tags_file"]
        )
        with open(tags_path, "w", encoding="utf-8") as fp:
            fp.write(TAG_FILE_MESSAGE)
            store_yaml_stream(fp, tag_data)

    # Write Galaxy requirements.yml file
    galaxy_filename = os.path.join(
        app_ctx.extra["dest_data_dir"], app_ctx.extra["galaxy_file"]
    )
    write_galaxy_requirements(galaxy_filename, dependency_data.deps)

    return 0


def rebuild_single_command() -> int:
    app_ctx = app_context.app_ctx.get()
    lib_ctx = app_context.lib_ctx.get()

    deps_filename = os.path.join(app_ctx.extra["data_dir"], app_ctx.extra["deps_file"])
    deps_file = DepsFile(deps_filename)
    dependency_data = deps_file.parse()
    python_requires: str | None
    try:
        python_requires = _extract_python_requires(
            PypiVer(dependency_data.ansible_core_version), dependency_data.deps
        )
    except ValueError:
        python_requires = None

    # Determine included collection versions
    ansible_core_version = PypiVer(dependency_data.ansible_core_version)
    included_versions = {
        collection: SemVer(version)
        for collection, version in dependency_data.deps.items()
    }

    galaxy_context = asyncio.run(create_galaxy_context())

    with tempfile.TemporaryDirectory() as tmp_dir:
        download_dir = os.path.join(tmp_dir, "collections")
        os.mkdir(download_dir, mode=0o700)

        # Download included collections
        asyncio.run(
            download_collections(
                included_versions,
                download_dir=download_dir,
                galaxy_context=galaxy_context,
                collection_cache=lib_ctx.collection_cache,
            )
        )

        # Get Ansible changelog, add new release
        ansible_changelog = ChangelogData.ansible(
            app_ctx.extra["data_dir"], app_ctx.extra["dest_data_dir"]
        )

        # Get changelog and porting guide data
        changelog = get_changelog(
            app_ctx.extra["ansible_version"],
            deps_dir=app_ctx.extra["data_dir"],
            deps_data=[dependency_data],
            collection_cache=lib_ctx.collection_cache,
            ansible_changelog=ansible_changelog,
            galaxy_context=galaxy_context,
        )

        # Create package and collections directories
        package_dir = os.path.join(
            tmp_dir, f'ansible-{app_ctx.extra["ansible_version"]}'
        )
        os.mkdir(package_dir, mode=0o700)
        ansible_collections_dir = os.path.join(package_dir, "ansible_collections")
        os.mkdir(ansible_collections_dir, mode=0o700)

        # Write the ansible release info to the collections dir
        write_release_py(app_ctx.extra["ansible_version"], ansible_collections_dir)

        # Write the ansible-community CLI program
        write_ansible_community_py(
            app_ctx.extra["ansible_version"], ansible_collections_dir
        )

        # Install collections
        collections_to_install = [
            p
            for f in os.listdir(download_dir)
            if os.path.isfile(p := os.path.join(download_dir, f))
        ]

        asyncio.run(install_together(collections_to_install, ansible_collections_dir))

        # Compose and write release notes to destination directory
        release_notes = ReleaseNotes.build(changelog)
        release_notes.write_changelog_to(package_dir)
        release_notes.write_porting_guide_to(package_dir)

        # Write changelog and porting guide also to destination directory
        release_notes.write_changelog_to(app_ctx.extra["dest_data_dir"])
        release_notes.write_porting_guide_to(app_ctx.extra["dest_data_dir"])

        # Write build scripts and files
        tags_path: str | None = None
        if app_ctx.extra["tags_file"]:
            tags_path = os.path.join(
                app_ctx.extra["data_dir"], app_ctx.extra["tags_file"]
            )

        write_all_build_files(
            package_dir=package_dir,
            collections_dir=ansible_collections_dir,
            ansible_version=app_ctx.extra["ansible_version"],
            dependency_data=dependency_data,
            ansible_core_version=ansible_core_version,
            python_requires=python_requires,
            tags_path=tags_path,
            debian=app_ctx.extra["debian"],
            sdist_src_dir=app_ctx.extra.get("sdist_src_dir"),
            ansible_core_checkout=asyncio.run(
                _get_ansible_core_path(tmp_dir, ansible_core_version)
            ),
            release_notes=release_notes,
        )

        # Check dependencies
        dep_errors = check_collection_dependencies(
            os.path.join(package_dir, "ansible_collections")
        )

        if dep_errors:
            is_error = app_ctx.extra["ansible_version"] >= PypiVer("6.3.0")
            warning_error = "ERROR" if is_error else "WARNING"
            print(f"{warning_error}: found collection dependency errors!")
            for error in dep_errors:
                print(f"{warning_error}: {error}")
            if is_error:
                return 3

        # Create source distribution
        make_dist_with_wheels(package_dir, app_ctx.extra["sdist_dir"])

    return 0


def generate_package_files_command() -> int:
    """
    PRIVATE, INTERNAL command to (re)generate package configuration files
    """
    app_ctx = app_context.app_ctx.get()
    package_dir = app_ctx.extra["package_dir"]
    tags_path: str | None = None
    if app_ctx.extra["tags_file"]:
        tags_path = os.path.join(app_ctx.extra["data_dir"], app_ctx.extra["tags_file"])

    deps_filename = os.path.join(app_ctx.extra["data_dir"], app_ctx.extra["deps_file"])
    deps_file = DepsFile(deps_filename)
    dependency_data = deps_file.parse()
    ansible_core_version = PypiVer(dependency_data.ansible_core_version)

    python_requires: str | None
    try:
        python_requires = _extract_python_requires(
            PypiVer(dependency_data.ansible_core_version), dependency_data.deps
        )
    except ValueError:
        python_requires = None

    with tempfile.TemporaryDirectory() as tmp_dir:
        write_all_build_files(
            package_dir=package_dir,
            collections_dir=app_ctx.extra["collections_dir"],
            ansible_version=app_ctx.extra["ansible_version"],
            dependency_data=dependency_data,
            ansible_core_version=ansible_core_version,
            python_requires=python_requires,
            tags_path=tags_path,
            debian=app_ctx.extra["debian"],
            ansible_core_checkout=asyncio.run(
                _get_ansible_core_path(tmp_dir, ansible_core_version)
            ),
        )

    return 0



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>

"""
Validate that collections tag their releases in their respective git repositories
"""

from __future__ import annotations

import asyncio
import os
import re
import sys
from collections.abc import AsyncGenerator, Collection
from typing import TYPE_CHECKING, TextIO, TypedDict

import asyncio_pool  # type: ignore[import]
from antsibull_core import app_context
from antsibull_core.dependency_files import DepsFile
from antsibull_core.logging import log
from antsibull_core.yaml import load_yaml_file, store_yaml_file

from antsibull.collection_meta import CollectionMetadata, CollectionsMetadata

if TYPE_CHECKING:
    from typing_extensions import NotRequired

TAG_MATCHER: re.Pattern[str] = re.compile(r"^.*refs/tags/(.*)$")
TAG_VERSION_REGEX: re.Pattern[str] = re.compile(r"^v?(.*)$")
mlog = log.fields(mod=__name__)


def validate_tags_file_command() -> int:
    """
    Ensure that collection versions in an Ansible release are tagged in
    collections' respective git repositories. This validates the tag file
    generated by the 'validate-tags' subcommand.
    """
    app_ctx = app_context.app_ctx.get()
    ignores = _get_ignores(app_ctx.extra["ignore"], app_ctx.extra["ignores_file"])
    tag_data = load_yaml_file(app_ctx.extra["tags_file"])
    return _print_validation_errors(
        tag_data,
        ignores,
        app_ctx.extra["error_on_useless_ignores"],
    )


def validate_tags_command() -> int:
    """
    Ensure that collection versions in an Ansible release are tagged in
    collections' respective git repositories.
    This retrieves a list of tags from the repository of each collection,
    optionally saves the retrieved tag data to a file,
    and then verifies the data.
    """
    app_ctx = app_context.app_ctx.get()
    ignores = _get_ignores(app_ctx.extra["ignore"], app_ctx.extra["ignores_file"])
    tag_data = asyncio.run(
        get_collections_tags(app_ctx.extra["data_dir"], app_ctx.extra["deps_file"])
    )
    if app_ctx.extra["output"]:
        store_yaml_file(app_ctx.extra["output"], tag_data)
    return _print_validation_errors(
        tag_data,
        ignores,
        app_ctx.extra["error_on_useless_ignores"],
    )


def _print_validation_errors(
    tag_data: dict[str, CollectionTagData],
    ignores: Collection[str] = (),
    error_on_useless_ignores: bool = True,
) -> int:
    """
    This takes the tag_data and prints any validation errors to stderr.
    """
    errors = validate_tags(tag_data, ignores, error_on_useless_ignores)
    if not errors:
        return 0
    for error in errors:
        print(error, file=sys.stderr)
    return 1


def _get_ignores(ignores: Collection[str], ignore_fp: TextIO | None) -> set[str]:
    ignores = set(ignores)
    if ignore_fp:
        ignores.update(
            line.strip()
            for line in ignore_fp
            if line.strip() and not line.startswith("#")
        )
    return ignores


##############
# Library code
##############


def validate_tags(
    tag_data: dict[str, CollectionTagData],
    ignores: Collection[str] = (),
    error_on_useless_ignores: bool = True,
) -> list[str]:
    """
    Validate that each collection in tag_data has a repository and a tag
    associated with it. Return a list of validation errors.

    :param tag_data: A tag data dictionary as returned by `get_collections_tags`
    :param ignores: A list of collection names for which to ignore errors
    :param error_on_useless_ignores: Whether to error for useless ignores
    """
    errors = []
    ignore_set = set(ignores)
    for name, data in tag_data.items():
        version = data["version"]
        if name in ignore_set:
            ignore_set.remove(name)
            if data["repository"] and data["tag"] and error_on_useless_ignores:
                errors.append(
                    f"useless ignore {name!r}: {name} {version} is properly tagged"
                )
        elif not data["repository"]:
            errors.append(
                f"{name}'s repository is not specified at all in collection-meta.yaml"
            )
        elif not data["tag"]:
            errors.append(f'{name} {version} is not tagged in {data["repository"]}')
    if ignore_set and error_on_useless_ignores:
        for name in ignore_set:
            errors.append(
                f"invalid ignore {name!r}: {name} does not match any collection"
            )
    return errors


class CollectionTagData(TypedDict):
    version: str
    repository: str | None
    tag: str | None
    collection_directory: NotRequired[str]


async def get_collections_tags(
    data_dir: str, deps_filename: str
) -> dict[str, CollectionTagData]:
    """
    Iterate over the collections in a CollectionsMetadata file,
    retrieve their tags, and return a dictionary
    of collection names mapped to dictionaries containing
    each collection's 'version' and 'repository' from a DepsFile
    and the 'tag' that matches.
    """
    lib_ctx = app_context.lib_ctx.get()

    deps_filename = os.path.join(data_dir, deps_filename)
    deps_data = DepsFile(deps_filename).parse()
    meta_data = CollectionsMetadata(data_dir)

    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        collection_tags = {}
        for name, data in meta_data.data.items():
            collection_tags[name] = pool.spawn_n(
                _get_collection_tags(deps_data.deps[name], data, name)
            )
        collection_tags = {name: await data for name, data in collection_tags.items()}
        return collection_tags


async def _get_collection_tags(
    version: str, meta_data: CollectionMetadata, name: str
) -> CollectionTagData:
    flog = mlog.fields(func="_get_collection_tags")
    repository = meta_data.repository
    data: CollectionTagData = {
        "version": version,
        "repository": repository,
        "tag": None,
    }
    if meta_data.collection_directory:
        data["collection_directory"] = meta_data.collection_directory
    if not repository:
        flog.debug("'repository' is None. Exitting...")
        return data
    tag_version_regex: re.Pattern[str] | None = None
    if meta_data.tag_version_regex:
        try:
            tag_version_regex = re.compile(meta_data.tag_version_regex)
        except re.error as err:
            flog.fields(err=err, collection=name).error(
                f"{tag_version_regex} is an invalid regex"
            )
            return data
    async for tag in _get_tags(repository):
        if _normalize_tag(tag, tag_version_regex) == version:
            data["tag"] = tag
            break
    return data


async def _get_tags(repository) -> AsyncGenerator[str, None]:
    flog = mlog.fields(func="_get_tags")
    args = (
        "git",
        "ls-remote",
        "--refs",
        "--tags",
        repository,
    )
    flog.debug(f"Running {args}")
    proc = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        # This makes it so git doesn't ask for a password when a repository
        # is inaccessible.
        env={"GIT_TERMINAL_PROMPT": "0"},
    )
    stdout, stderr = await proc.communicate()
    flog.fields(stderr=stderr, returncode=proc.returncode).debug("Ran git ls-remote")
    if proc.returncode != 0:
        flog.error(f"Failed to fetch tags for {repository}")
        return
    tags: list[str] = stdout.decode("utf-8").splitlines()
    if not tags:
        flog.warning(f"{repository} does not have any tags")
        return
    for tag in tags:
        match = TAG_MATCHER.match(tag)
        if match:
            yield match.group(1)
        else:
            flog.debug(f"git ls-remote output line skipped: {tag}")


def _normalize_tag(tag: str, regex: re.Pattern[str] | None) -> str | None:
    regex = regex or TAG_VERSION_REGEX
    if match := regex.match(tag):
        return match.group(1)
    return tag



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2021
"""Check collection dependencies."""

from __future__ import annotations

import json
import pathlib
from collections import namedtuple
from collections.abc import Mapping

from antsibull_core import app_context
from semantic_version import SimpleSpec as SemVerSpec
from semantic_version import Version as SemVer

CollectionRecord = namedtuple("CollectionRecord", ("version", "dependencies"))


def parse_manifest(collection_dir: pathlib.Path) -> Mapping[str, CollectionRecord]:
    """Parse MANIFEST.json for a collection."""
    manifest = collection_dir.joinpath("MANIFEST.json")
    with manifest.open() as f:
        manifest_data = json.load(f)["collection_info"]

    collection_record = {
        f'{manifest_data["namespace"]}.{manifest_data["name"]}': CollectionRecord(
            manifest_data["version"], manifest_data["dependencies"]
        )
    }

    return collection_record


def analyze_deps(collections: Mapping[str, CollectionRecord]) -> list[str]:
    """Analyze dependencies of a set of collections. Return list of errors found."""
    errors = []

    # Look at dependencies
    # make sure their dependencies are found
    for collection_name, collection_info in collections.items():
        for dep_name, dep_version_spec in collection_info.dependencies.items():
            if dep_name not in collections:
                errors.append(
                    f"{collection_name} missing: {dep_name} ({dep_version_spec})"
                )
                continue

            dependency_version = SemVer(collections[dep_name].version)
            if dependency_version not in SemVerSpec(dep_version_spec):
                errors.append(
                    f"{collection_name} {collection_info.version} version_conflict:"
                    f" {dep_name}-{dependency_version} but needs"
                    f" {dep_version_spec}"
                )
                continue

    return errors


def check_collection_dependencies(collection_root: str) -> list[str]:
    """Analyze dependencies between collections in a collection root."""
    ansible_collection_dir = pathlib.Path(collection_root)
    errors = []

    collections: dict[str, CollectionRecord] = {}
    for namespace_dir in (n for n in ansible_collection_dir.iterdir() if n.is_dir()):
        for collection_dir in (c for c in namespace_dir.iterdir() if c.is_dir()):
            try:
                collections.update(parse_manifest(collection_dir))
            except FileNotFoundError:
                errors.append(f"{collection_dir} is not a valid collection")

    errors.extend(analyze_deps(collections))
    return errors


def validate_dependencies_command() -> int:
    """CLI functionality for analyzing dependencies."""
    app_ctx = app_context.app_ctx.get()

    collection_root: str = app_ctx.extra["collection_root"]

    errors = check_collection_dependencies(collection_root)

    for error in errors:
        print(error)

    return 3 if errors else 0



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Code for generating Python build configuration for ansible
"""

from __future__ import annotations

import copy
import os
from collections import defaultdict
from collections.abc import Collection, Iterator
from configparser import ConfigParser
from email import message_from_file
from email.message import Message
from pathlib import Path
from typing import Any, Union

from antsibull_core.dependency_files import DependencyFileData
from jinja2 import Template
from packaging.version import Version as PypiVer

from .constants import (
    ANSIBLE_FORUM_URL,
    BUILD_DATA_URL,
    COLLECTION_EXCLUDE_DIRS,
    DOCSITE_BASE_URL,
    DOCSITE_COMMUNITY_URL,
    MINIMUM_ANSIBLE_VERSIONS,
)
from .utils.get_pkg_data import get_antsibull_data


class IniType:
    """
    Python collections with custom __str__ methods that
    can be serialized as INI
    """

    def _i_iter_lines(self) -> Iterator[str]:
        raise NotImplementedError

    def __str__(self) -> str:
        return "\n".join(self._i_iter_lines())


class IniDict(IniType, dict):
    def _i_iter_lines(self) -> Iterator[str]:
        yield ""
        for key, value in self.items():
            if key == "":
                yield f"={value}"
            else:
                yield f"{key} = {value}"


class IniList(IniType, list):
    def _i_iter_lines(self) -> Iterator[str]:
        yield ""
        for value in self:
            yield f"  {value}"


INI_TYPES = Union[IniType, str, bool]

OLD_URLS = IniDict(
    {
        "Bug Tracker": "https://github.com/ansible/ansible/issues",
        "Code of Conduct": DOCSITE_COMMUNITY_URL + "/code_of_conduct.html",
        "Documentation": DOCSITE_BASE_URL,
        "Mailing lists": DOCSITE_COMMUNITY_URL
        + "/communication.html#mailing-list-information",
        "Source Code": "https://github.com/ansible/ansible",
    }
)

NEW_URLS = IniDict(
    {
        "Build Data": BUILD_DATA_URL,
        "Code of Conduct": DOCSITE_COMMUNITY_URL + "/code_of_conduct.html",
        "Documentation": DOCSITE_BASE_URL,
        "Forum": ANSIBLE_FORUM_URL,
        "Mailing lists": DOCSITE_COMMUNITY_URL
        + "/communication.html#mailing-list-information",
    }
)

DEFAULT_METADATA: dict[str, INI_TYPES] = {
    "name": "ansible",
    "description": "Radically simple IT automation",
    "long_description": "file: README.rst",
    "long_description_content_type": "text/x-rst",
    "author": "Ansible, Inc.",
    "author_email": "info@ansible.com",
    "url": "https://ansible.com/",
    "license": "GPL-3.0-or-later",
    "classifiers": IniList(
        [
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Framework :: Ansible",
            "Intended Audience :: Developers",
            "Intended Audience :: Information Technology",
            "Intended Audience :: System Administrators",
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Natural Language :: English",
            "Operating System :: POSIX",
        ]
    ),
}

DEFAULT_OPTIONS: dict[str, INI_TYPES] = {"zip_safe": False}
DEFAULT_ENTRY_POINTS: dict[str, INI_TYPES] = {
    "console_scripts": IniDict(
        {"ansible-community": "ansible_collections.ansible_community:main"}
    )
}

DEFAULT_CONFIG: dict[str, dict[str, INI_TYPES]] = {
    "metadata": DEFAULT_METADATA,
    "options": DEFAULT_OPTIONS,
    "options.entry_points": DEFAULT_ENTRY_POINTS,
}


def _get_collection_data(collection_path: Path) -> list[str]:
    directories = []
    for root, dirs, _ in os.walk(collection_path, topdown=True, followlinks=True):
        if root == str(collection_path):
            # Make sure that all directories starting with '.', and all
            # directories called 'tests' or 'docs', are not traversed into.
            for dirname in list(dirs):
                if dirname in COLLECTION_EXCLUDE_DIRS or dirname.startswith("."):
                    dirs.remove(dirname)
            continue
        relative_dir = str(Path(root).relative_to(collection_path))
        directories.append(relative_dir)
    return sorted(directories)


def _get_exclude_package_data(
    collection_names: Collection[str], collection_root: Path
) -> Iterator[str]:
    for collection in collection_names:
        relpath = Path(collection.replace(".", "/"))
        path = collection_root / relpath
        for directory in COLLECTION_EXCLUDE_DIRS:
            if path.joinpath(directory).is_dir():
                yield str(relpath / directory / "*")
        if any(file.name.startswith(".") for file in path.iterdir()):
            yield str(relpath / ".*")


class BuildMetaMaker:
    """
    Generate a setup.cfg and other Python metadata for ansible
    """

    package_dir: Path
    collections_dir: Path
    ansible_version: PypiVer
    dependency_data: DependencyFileData
    ansible_core_version: PypiVer
    ansible_core_checkout: Path
    python_requires: str | None

    config: defaultdict[str, dict[str, INI_TYPES]]
    ansible_core_metadata: Message
    _collection_directories: dict[str, list[str]]

    def __init__(
        self,
        *,
        package_dir: str | os.PathLike[str],
        collections_dir: str | os.PathLike[str] | None = None,
        ansible_version: PypiVer,
        dependency_data: DependencyFileData,
        ansible_core_version: PypiVer,
        ansible_core_checkout: str | os.PathLike[str],
        initial_config: dict[str, dict[str, INI_TYPES]] | None = None,
        python_requires: str | None,
    ) -> None:
        self.package_dir = Path(package_dir)
        if collections_dir:
            self.collections_dir = Path(collections_dir)
        else:
            self.collections_dir = self.package_dir / "ansible_collections"
        self.ansible_version = ansible_version
        self.dependency_data = dependency_data
        self.ansible_core_version = ansible_core_version
        self.ansible_core_checkout = Path(ansible_core_checkout)
        initial_config = copy.deepcopy(initial_config or DEFAULT_CONFIG)
        self.config = defaultdict(dict, initial_config)
        self.python_requires = python_requires

        with (self.ansible_core_checkout / "PKG-INFO").open(encoding="utf-8") as fp:
            self.ansible_core_metadata = message_from_file(fp)

        self._generated = False
        # TODO: Remove once we drop LegacyBuildMetaMaker
        self._collection_directories: dict[str, str] = {}

    def __getitem__(self, key: str) -> Any:
        return self.config[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.config[key] = value

    @property
    def core_python_requires(self) -> str:
        result = self.ansible_core_metadata.get("Requires-Python")
        if not result:
            raise ValueError("ansible-core's metadata is missing 'Requires-Python'")
        return result

    @property
    def core_python_classifiers(self) -> list[str]:
        result = self.ansible_core_metadata.get_all("Classifier")
        if not result:
            raise ValueError("ansible-core's metadata is missing Classifiers")
        return [
            classifier
            for classifier in result
            if classifier.startswith("Programming Language :: Python ::")
        ]

    def add_collection_ignores(self) -> None:
        if self.ansible_version >= MINIMUM_ANSIBLE_VERSIONS["PACKAGE_DATA_NEW_METHOD"]:
            self._package_data_new_method()
        else:
            self._package_data_old_method()

    def _package_data_new_method(self) -> None:
        collection_namespaces: dict[str, list[str]] = defaultdict(list)
        # We only need this as an attribute for the LegacyBuildMetaMaker compat
        # TODO: Remove once we drop LegacyBuildMetaMaker
        collection_directories = self._collection_directories
        for collection in self.dependency_data.deps:
            namespace, name = collection.split(".", 1)
            collection_namespaces[namespace].append(name)
            collection_path = self.collections_dir / namespace / name
            collection_directories[collection] = _get_collection_data(collection_path)

        self["options"]["package_dir"] = IniDict({"": "."})
        self["options"]["packages"] = "find_namespace:"

        self["options.packages.find"]["include"] = IniList(
            ["ansible_collections", "ansible_collections.*"]
        )
        self["options.packages.find"]["exclude"] = IniList()
        excludes = self["options.packages.find"]["exclude"]

        for collection in sorted(self.dependency_data.deps):
            excludes.extend(
                (
                    f"ansible_collections.{collection}.tests",
                    f"ansible_collections.{collection}.tests.*",
                    f"ansible_collections.{collection}.docs",
                    f"ansible_collections.{collection}.docs.*",
                )
            )
            data = self["options.package_data"][f"ansible_collections.{collection}"] = (
                IniList()
            )

            data.append("*")
            for directory in collection_directories[collection]:
                data.extend([f"{directory}/*", f"{directory}/.*"])

    def _package_data_old_method(self) -> None:
        self["options"]["packages"] = IniList(["ansible_collections"])
        self["options"]["include_package_data"] = True
        self["options.exclude_package_data"]["ansible_collections"] = IniList(
            _get_exclude_package_data(self.dependency_data.deps, self.collections_dir)
        )

    def generate(self) -> None:
        self["metadata"]["version"] = self.ansible_version
        self["metadata"].setdefault("classifiers", IniList()).extend(
            self.core_python_classifiers
        )
        self["metadata"].setdefault(
            "project_urls",
            (
                NEW_URLS
                if self.ansible_version
                >= MINIMUM_ANSIBLE_VERSIONS["BUILD_META_NEW_URLS"]
                else OLD_URLS
            ),
        )
        self["options"].setdefault("install_requires", IniList()).append(
            f"ansible-core ~= {self.ansible_core_version}"
        )

        self["options"]["python_requires"] = (
            self.python_requires or self.core_python_requires
        )

        self.add_collection_ignores()

        self._generated = True

    def write(self, file: Path | None = None) -> None:
        if not self._generated:
            self.generate()
        parser = ConfigParser()
        parser.read_dict(self.config)
        file = file or self.package_dir.joinpath("setup.cfg")
        with file.open("w", encoding="utf-8") as fp:
            parser.write(fp)


class LegacyBuildMetaMaker:
    """
    Generate a setup.py and other Python metadata for ansible.
    This is a wrapper around `BuildMetaMaker`.
    """

    def __init__(
        self,
        *,
        package_dir: str | os.PathLike[str],
        collections_dir: str | os.PathLike[str] | None = None,
        ansible_version: PypiVer,
        dependency_data: DependencyFileData,
        ansible_core_version: PypiVer,
        ansible_core_checkout: str | os.PathLike[str],
        python_requires: str | None,
    ) -> None:
        self.maker: BuildMetaMaker = BuildMetaMaker(
            package_dir=package_dir,
            collections_dir=collections_dir,
            ansible_version=ansible_version,
            dependency_data=dependency_data,
            ansible_core_version=ansible_core_version,
            ansible_core_checkout=ansible_core_checkout,
            initial_config=DEFAULT_CONFIG,
            python_requires=python_requires,
        )

    def write(self, file: Path | None = None) -> None:
        file = file or self.maker.package_dir / "setup.py"

        self.maker.generate()

        collection_exclude_paths = self.maker["options.exclude_package_data"].get(
            "ansible_collections", []
        )

        setup_tmpl = Template(get_antsibull_data("ansible-setup_py.j2").decode("utf-8"))
        setup_contents = setup_tmpl.render(
            version=self.maker.ansible_version,
            ansible_core_package_name="ansible-core",
            ansible_core_version=self.maker.ansible_core_version,
            collection_deps="",
            # not PACKAGE_DATA_NEW_METHOD
            collection_exclude_paths=sorted(collection_exclude_paths),
            # PACKAGE_DATA_NEW_METHOD
            collection_names=sorted(self.maker.dependency_data.deps),
            # pylint: disable-next=protected-access
            collection_directories=self.maker._collection_directories,
            #
            python_requires=self.maker["options"]["python_requires"],
            PypiVer=PypiVer,
        )
        file.write_text(setup_contents, encoding="utf-8")



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Errors related to the from_source package
"""

from __future__ import annotations


class CloneError(Exception):
    """
    An issue occured with cloning a collection's upstream repository
    """

    def __init__(self, message: str, collection: str) -> None:
        self.message = message
        self.collection = collection
        super().__init__(f"{self.collection}: {self.message}")


__all__ = ("CloneError",)



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Misc. utility functions for the `from_source` package
"""


from __future__ import annotations

from collections.abc import Sequence
from fnmatch import fnmatch

from semantic_version import Version as SemVer

from antsibull.tagging import CollectionTagData
from antsibull.types import CollectionName


def filter_tag_data(
    collections: dict[CollectionName, CollectionTagData],
    globs: Sequence[str] | None = None,
) -> dict[CollectionName, CollectionTagData]:
    return {
        collection: data
        for collection, data in collections.items()
        if globs is None or any(fnmatch(collection, glob) for glob in globs)
    }


def tag_data_as_version_mapping(
    collections: dict[CollectionName, CollectionTagData]
) -> dict[CollectionName, SemVer]:
    """
    Convert a collection: CollectionTagData mapping to a collection: version mapping
    """
    return {
        collection: SemVer(data["version"]) for collection, data in collections.items()
    }


__all__ = ("filter_tag_data", "tag_data_as_version_mapping")



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import argparse
import asyncio
import contextlib
import dataclasses
import os
import shutil
import sys
import tarfile
import tempfile
import zipfile
from collections.abc import Iterator, MutableMapping, Sequence
from pathlib import Path
from subprocess import run
from typing import Any, Union

import aiofiles
import aiohttp
from packaging.version import Version as PypiVer

import antsibull.build_ansible_commands
from antsibull.cli import antsibull_build
from antsibull.constants import MINIMUM_ANSIBLE_VERSIONS
from antsibull.utils.paths import temp_or_dir

HERE = Path(__file__).resolve().parent
ROOT = HERE.parent

DEFAULT_CACHE_DIR = HERE / ".cache"
DEFAULT_PACKAGE_DIR = HERE / "test_data" / "package-files"

PYPI_PATH = "https://files.pythonhosted.org/packages"
WHEEL_PATH = f"{PYPI_PATH}/py3/a/ansible/ansible-{{version}}-py3-none-any.whl"
SDIST_PATH = f"{PYPI_PATH}/source/a/ansible/ansible-{{version}}.tar.gz"

ANTSIBULL_BUILD = os.environ.get("ANTSIBULL_BUILD", "antsibull-build")
PLACEHOLDER_ANTSIBULL_VERSION = "(ANTSIBULL_VERSION)"


@dataclasses.dataclass
class AnsibleSdist:
    version: str
    dest_dir: Path

    @property
    def nv(self) -> str:
        return f"ansible-{self.version}"

    @property
    def name(self) -> Path:
        return Path(self.nv + ".tar.gz")

    @property
    def dest(self) -> Path:
        return self.dest_dir / self.name

    @property
    def url(self) -> str:
        return "/".join((PYPI_PATH, "source", "a", "ansible", str(self.name)))

    def extract_collections(self, extract_dir: Path) -> None:
        with tarfile.TarFile.open(self.dest) as file:
            collections_dir = str(Path(self.nv, "ansible_collections")) + "/"
            members: list[tarfile.TarInfo] = []
            for member in file.getmembers():
                if not member.name.startswith(collections_dir):
                    continue
                member.path = member.name[len(self.nv) + 1 :]
                members.append(member)
            file.extractall(extract_dir, members)

    def list_files(self) -> list[str]:
        with tarfile.TarFile.open(self.dest) as file:
            files: list[str] = []
            for info in file.getmembers():
                if info.isdir():
                    files.append(info.name + "/")
                else:
                    files.append(info.name)
            return sorted(files)


@dataclasses.dataclass
class AnsibleWheel:
    version: str
    dest_dir: Path

    @property
    def nv(self) -> str:
        return f"ansible-{self.version}"

    @property
    def name(self) -> Path:
        return Path(f"{self.nv}-py3-none-any.whl")

    @property
    def dest(self) -> Path:
        return self.dest_dir / self.name

    @property
    def url(self) -> str:
        return "/".join((PYPI_PATH, "py3", "a", "ansible", str(self.name)))

    def list_files(self) -> list[str]:
        with zipfile.ZipFile(self.dest) as file:
            files: list[str] = []
            for info in file.infolist():
                if info.is_dir():
                    files.append(info.filename + "/")
                else:
                    files.append(info.filename)
            return sorted(files)


DIST_TUPLE = tuple[Union[AnsibleSdist, AnsibleWheel], ...]


async def download_file(session: aiohttp.ClientSession, url: str, dest: Path) -> None:
    print(f"Downloading {url}...")
    async with session.get(url) as resp:
        async with aiofiles.open(dest, "wb") as fp:
            while chunk := await resp.content.read(4096):
                await fp.write(chunk)


def download_command(*, version: str, cache_dir: Path, force_dl: bool) -> None:
    async def dl() -> None:
        async with aiohttp.ClientSession() as session:
            dist = AnsibleSdist(version, cache_dir)
            if force_dl or not dist.dest.exists():
                await download_file(session, dist.url, dist.dest)

    asyncio.run(dl())


def generate_package_files(
    version: str,
    cached_dist: AnsibleSdist,
    extract_dir: Path,
    data_dir: Path,
    force_generate_setup_cfg: bool,
) -> None:
    cached_dist.extract_collections(extract_dir)
    cm = (
        patch_dict(MINIMUM_ANSIBLE_VERSIONS, "BUILD_META_MAKER", PypiVer(version))
        if force_generate_setup_cfg
        else contextlib.nullcontext()
    )
    cm2 = (
        patch_dict(MINIMUM_ANSIBLE_VERSIONS, "BUILD_META_NEW_URLS", PypiVer(version))
        if force_generate_setup_cfg
        else contextlib.nullcontext()
    )
    with (
        cm,
        cm2,
        patch_object(
            antsibull.build_ansible_commands,
            "antsibull_version",
            PLACEHOLDER_ANTSIBULL_VERSION,
        ),
    ):
        if r := antsibull_build.run(
            [
                "antsibull-build",
                "generate-package-files",
                f"--data-dir={data_dir}",
                "--tags-file",
                f"--package-dir={extract_dir}",
                version,
            ]
        ):
            sys.exit(r)


@contextlib.contextmanager
def patch_dict(mapping: MutableMapping, key: Any, value: Any) -> Iterator[None]:
    old = mapping[key]
    try:
        mapping[key] = value
        yield
    finally:
        mapping[key] = old


@contextlib.contextmanager
def patch_object(object: Any, attr: str, new_value: Any) -> Iterator[None]:
    old_value = getattr(object, attr)
    try:
        setattr(object, attr, new_value)
        yield
    finally:
        setattr(object, attr, old_value)


def write_file_list(
    version: str, source_dir: Path, build_dir: Path | None = None
) -> DIST_TUPLE:
    with temp_or_dir(build_dir) as build_dir:
        run(
            [
                sys.executable,
                "-m",
                "build",
                f"--outdir={build_dir}",
                "--config-setting=--quiet",
                source_dir,
            ],
            check=True,
        )
        dists: DIST_TUPLE = (
            AnsibleSdist(version, Path(build_dir)),
            AnsibleWheel(version, Path(build_dir)),
        )
        dist_list_dir = source_dir / "dist-files"
        dist_list_dir.mkdir(exist_ok=True)
        for dist in dists:
            dist_list = dist_list_dir / f"{dist.name}.txt"
            dist_list.write_text("\n".join(dist.list_files()) + "\n")
        return dists


def check_command(
    *,
    version: str,
    package_dir: Path,
    cache_dir: Path,
    data_dir: Path,
    build_check: Path,
    build_dir: Path | None,
    force_generate_setup_cfg: bool,
) -> None:
    package_dir = package_dir / version
    cached_dist = AnsibleSdist(version, cache_dir)

    with temp_or_dir() as extract_dir:
        generate_package_files(
            version, cached_dist, extract_dir, data_dir, force_generate_setup_cfg
        )
        diff_args = ["-ur", "-x", "ansible_collections", "-x", "*egg-info"]
        if build_check:
            write_file_list(version, extract_dir, build_dir)
        else:
            diff_args.extend(("-x", "dist-files"))
        run(["diff", *diff_args, package_dir, extract_dir], check=True)

        # Make sure files can be regenerated
        generate_package_files(
            version, cached_dist, extract_dir, data_dir, force_generate_setup_cfg
        )
        run(["diff", *diff_args, package_dir, extract_dir], check=True)


def regen_command(
    *,
    version: str,
    package_dir: Path,
    cache_dir: Path,
    data_dir: Path,
    build_check: Path,
    clean: bool,
    build_dir: Path | None,
    force_generate_setup_cfg: bool,
) -> None:
    package_dir = package_dir / version
    cached_dist = AnsibleSdist(version, cache_dir)

    if clean:
        shutil.rmtree(package_dir, True)
    package_dir.mkdir(exist_ok=True)

    generate_package_files(
        version, cached_dist, package_dir, data_dir, force_generate_setup_cfg
    )
    if build_check:
        write_file_list(version, package_dir, build_dir)
    shutil.rmtree(package_dir / "ansible_collections", True)
    if egg_info := next(package_dir.glob("*.egg-info"), None):
        shutil.rmtree(egg_info, True)


def parse_args(argv: Sequence[str] | None = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(title="action", dest="action", required=True)

    download_parser = subparsers.add_parser("download")
    download_parser.set_defaults(function=download_command)
    download_parser.add_argument("version")
    download_parser.add_argument("--cache-dir", type=Path, default=DEFAULT_CACHE_DIR)
    download_parser.add_argument(
        "--force-dl", action=argparse.BooleanOptionalAction, default=False
    )

    _package_data_parser = argparse.ArgumentParser(add_help=False)
    _package_data_parser.add_argument("version")
    _package_data_parser.add_argument(
        "--package-dir", type=Path, default=DEFAULT_PACKAGE_DIR
    )
    _package_data_parser.add_argument(
        "--cache-dir", type=Path, default=DEFAULT_CACHE_DIR
    )
    _package_data_parser.add_argument("--data-dir", type=Path, required=True)
    _package_data_parser.add_argument(
        "--build-check",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Default: true",
    )
    _package_data_parser.add_argument(
        "--build-dir",
        help="Directory to use for storing temporary build artifacts"
        " when --build-check is used",
        type=Path,
    )
    _package_data_parser.add_argument(
        "--force-generate-setup-cfg",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Generate setup.cfg even if that's not the default for VERSION."
        " Default: False",
    )

    check_parser = subparsers.add_parser("check", parents=[_package_data_parser])
    check_parser.set_defaults(function=check_command)

    regen_parser = subparsers.add_parser("regen", parents=[_package_data_parser])
    regen_parser.set_defaults(function=regen_command)
    regen_parser.add_argument(
        "--clean",
        default=False,
        action=argparse.BooleanOptionalAction,
        help="Default: false",
    )

    return parser.parse_args(argv)


def main(argv: Sequence[str] | None = None):
    namespace = parse_args(argv)
    args = vars(namespace).copy()
    del args["action"]
    del args["function"]
    return namespace.function(**args)


if __name__ == "__main__":
    main()



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

from pathlib import Path

import pytest

HERE = Path(__file__).resolve().parent
TEST_DATA = HERE / "test_data"


@pytest.fixture
def test_data_path() -> Path:
    return TEST_DATA



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020
"""Entrypoint to the antsibull-build tool."""

from __future__ import annotations

import argparse
import os.path
import sys
from pathlib import Path

import twiggy  # type: ignore[import]
from antsibull_core.logging import initialize_app_logging, log
from packaging.version import Version as PypiVer

initialize_app_logging()

# We have to call initialize_app_logging() before these imports so that the log object is configured
# correctly before other antisbull modules make copies of it.
# pylint: disable=wrong-import-position,ungrouped-imports
from antsibull_core import app_context  # noqa: E402
from antsibull_core.args import (  # noqa: E402
    InvalidArgumentError,
    get_toplevel_parser,
    normalize_toplevel_options,
)
from antsibull_core.config import ConfigError, load_config  # noqa: E402
from antsibull_core.vendored._argparse_booleanoptionalaction import (  # noqa: E402
    BooleanOptionalAction,
)

from ..announcements import ACTIONS as ALLOWED_SEND_ACTIONS  # noqa: E402
from ..announcements import (  # noqa: E402
    announcements_command,
    send_announcements_command,
)
from ..build_ansible_commands import (  # noqa: E402
    build_single_command,
    generate_package_files_command,
    prepare_command,
    rebuild_single_command,
)
from ..build_changelog import build_changelog  # noqa: E402
from ..constants import MINIMUM_ANSIBLE_VERSION, SANITY_TESTS_DEFAULT  # noqa: E402
from ..dep_closure import validate_dependencies_command  # noqa: E402
from ..from_source import verify_upstream_command  # noqa: E402
from ..from_source.verify import LENIENT_FILE_ERROR_IGNORES, FileError  # noqa: E402
from ..new_ansible import new_ansible_command  # noqa: E402
from ..sanity_tests import sanity_tests_command  # noqa: E402
from ..tagging import validate_tags_command, validate_tags_file_command  # noqa: E402

# pylint: enable=wrong-import-position


mlog = log.fields(mod=__name__)

DEFAULT_FILE_BASE = "ansible"
DEFAULT_PIECES_FILE = f"{DEFAULT_FILE_BASE}.in"

ARGS_MAP = {
    "new-ansible": new_ansible_command,
    "prepare": prepare_command,
    "single": build_single_command,
    "changelog": build_changelog,
    "rebuild-single": rebuild_single_command,
    "validate-deps": validate_dependencies_command,
    "validate-tags": validate_tags_command,
    "validate-tags-file": validate_tags_file_command,
    "generate-package-files": generate_package_files_command,
    "verify-upstreams": verify_upstream_command,
    "sanity-tests": sanity_tests_command,
    "announcements": announcements_command,
    "send-announcements": send_announcements_command,
}
DISABLE_VERIFY_UPSTREAMS_IGNORES_SENTINEL = "NONE"
DEFAULT_ANNOUNCEMENTS_DIR = Path("build/announce")


def _normalize_commands(
    args: argparse.Namespace,  # pylint: disable=unused-argument
) -> None:
    # If command names change and old ones need to be deprecated, do that here.
    # Check out the git history for examples.
    pass


def _normalize_build_options(args: argparse.Namespace) -> None:
    if args.command in (
        "validate-deps",
        "validate-tags-file",
        "verify-upstreams",
        "sanity-tests",
        "send-announcements",
    ):
        return

    if args.ansible_version < MINIMUM_ANSIBLE_VERSION:
        raise InvalidArgumentError(
            f"Ansible < {MINIMUM_ANSIBLE_VERSION} is not supported"
            " by this antsibull version."
        )

    if not os.path.isdir(args.data_dir):
        raise InvalidArgumentError(f"{args.data_dir} must be an existing directory")


def _normalize_build_write_data_options(args: argparse.Namespace) -> None:
    if args.command not in (
        "new-ansible",
        "prepare",
        "single",
        "rebuild-single",
        "changelog",
    ):
        return

    if args.dest_data_dir is None:
        args.dest_data_dir = args.data_dir

    if not os.path.isdir(args.dest_data_dir):
        raise InvalidArgumentError(
            f"{args.dest_data_dir} must be an existing directory"
        )


def _normalize_new_release_options(args: argparse.Namespace) -> None:
    if args.command != "new-ansible":
        return

    if args.pieces_file is None:
        args.pieces_file = DEFAULT_PIECES_FILE

    pieces_path = os.path.join(args.data_dir, args.pieces_file)
    if not os.path.isfile(pieces_path):
        raise InvalidArgumentError(
            f"The pieces file, {pieces_path}, must already"
            " exist. It should contain one namespace.collection"
            " per line"
        )

    compat_version_part = f"{args.ansible_version.major}"

    if args.build_file is None:
        basename = os.path.basename(os.path.splitext(args.pieces_file)[0])
        args.build_file = f"{basename}-{compat_version_part}.build"

    if args.constraints_file is None:
        basename = os.path.basename(os.path.splitext(args.pieces_file)[0])
        args.constraints_file = f"{basename}-{compat_version_part}.constraints"


def _check_release_build_directories(args: argparse.Namespace) -> None:
    if args.command in ("single", "rebuild-single"):
        if not os.path.isdir(args.sdist_dir):
            raise InvalidArgumentError(
                f"{args.sdist_dir} must be an existing directory"
            )

    if args.command in ("rebuild-single",):
        if args.sdist_src_dir is not None and os.path.exists(args.sdist_src_dir):
            raise InvalidArgumentError(f"{args.sdist_src_dir} must not exist")


def _normalize_release_build_options(args: argparse.Namespace) -> None:  # noqa: C901
    deps_file_only: tuple[str, ...] = ("announcements",)
    if args.command not in (
        "prepare",
        "single",
        "rebuild-single",
        "generate-package-files",
        *deps_file_only,
    ):
        return

    compat_version_part = f"{args.ansible_version.major}"

    if args.build_file is None:
        args.build_file = DEFAULT_FILE_BASE + f"-{compat_version_part}.build"

    build_filename = os.path.join(args.data_dir, args.build_file)
    if args.command not in deps_file_only and not os.path.isfile(build_filename):
        raise InvalidArgumentError(
            f"The build file, {build_filename} must already exist."
            " It should contains one namespace.collection and range"
            " of versions per line"
        )

    if args.constraints_file is None:
        basename = os.path.basename(os.path.splitext(args.build_file)[0])
        args.constraints_file = f"{basename}.constraints"

    if args.deps_file is None:
        version_suffix = f"-{compat_version_part}"
        basename = os.path.basename(os.path.splitext(args.build_file)[0])
        if basename.endswith(version_suffix):
            basename = basename[: -len(version_suffix)]

        args.deps_file = f"{basename}-{args.ansible_version}.deps"

    if args.command in deps_file_only:
        return

    if args.tags_file:
        _check_tags_file(args)

    if args.command in ("prepare", "single") and args.galaxy_file is None:
        version_suffix = f"-{compat_version_part}"
        basename = os.path.basename(os.path.splitext(args.build_file)[0])
        if basename.endswith(version_suffix):
            basename = basename[: -len(version_suffix)]

        args.galaxy_file = f"{basename}-{args.ansible_version}.yaml"

    _check_release_build_directories(args)


def _check_tags_file(args: argparse.Namespace) -> None:
    if args.tags_file == "DEFAULT":
        args.tags_file = f"{DEFAULT_FILE_BASE}-{args.ansible_version}-tags.yaml"
    tags_path = os.path.join(args.data_dir, args.tags_file)

    if args.command == "rebuild-single" and not os.path.isfile(tags_path):
        raise InvalidArgumentError(f"{tags_path} does not exist!")


def _normalize_release_rebuild_options(args: argparse.Namespace) -> None:
    if args.command not in ("rebuild-single", "validate-tags"):
        return

    deps_filename = os.path.join(args.data_dir, args.deps_file)
    if not os.path.isfile(deps_filename):
        raise InvalidArgumentError(
            f"The dependency file, {deps_filename} must already exist."
        )


def _normalize_validate_tags_options(args: argparse.Namespace) -> None:
    if args.command not in ("validate-tags",):
        return
    if args.deps_file is None:
        args.deps_file = DEFAULT_FILE_BASE + f"-{args.ansible_version}.deps"


def _normalize_validate_tags_file_options(args: argparse.Namespace) -> None:
    if args.command not in ("validate-tags-file", "validate-upstreams"):
        return
    if not os.path.exists(args.tags_file):
        raise InvalidArgumentError(f"{args.tags_file} does not exist!")


def _normalize_generate_package_files_options(args: argparse.Namespace) -> None:
    if args.command not in ("generate-package-files",):
        return
    if not os.path.isdir(args.package_dir):
        raise InvalidArgumentError(f"{args.package_dir} does not exist!")

    args.collections_dir = args.collections_dir or os.path.join(
        args.package_dir, "ansible_collections"
    )
    if not os.path.isdir(args.collections_dir):
        raise InvalidArgumentError(f"{args.collections_dir} does not exist!")


def _normalize_verify_upstream_options(args: argparse.Namespace) -> None:
    if args.command not in ("verify-upstreams",):
        return
    if all((args.tree_dir, args.checkouts_dir)):
        tree_dir: Path = args.tree_dir
        checkouts_dir: Path = args.checkouts_dir
        if tree_dir.resolve() == checkouts_dir.resolve():
            raise InvalidArgumentError(
                "--tree-dir and --checkouts-dir must be unique values"
            )
    for arg, value in {
        "--tree-dir": args.tree_dir,
        "--checkouts-dir": args.checkouts_dir,
    }.items():
        if not value:
            continue
        if (directory := value / "ansible_collections").exists():
            raise InvalidArgumentError(f"{arg}: {directory} must not exist")
    if DISABLE_VERIFY_UPSTREAMS_IGNORES_SENTINEL in (args.ignores or set()):
        args.ignore = []
    if args.ignores is None:
        args.ignores = LENIENT_FILE_ERROR_IGNORES


def _normalize_announcements_options(args: argparse.Namespace) -> None:
    if args.command not in ("announcements",):
        return
    directory: Path = args.output_dir
    directory.mkdir(parents=True, exist_ok=True)


def _normalize_send_announcements_options(args: argparse.Namespace) -> None:
    if args.command not in ("send-announcements",):
        return
    directory: Path = args.announcements_dir
    if not args.announcements_dir / "announcements.json":
        raise InvalidArgumentError(
            f"'announcements.json' does not exist in directory {directory}!"
        )
    if args.send_actions is None:
        args.send_actions = set(ALLOWED_SEND_ACTIONS)


def parse_args(program_name: str, args: list[str]) -> argparse.Namespace:
    """
    Parse and coerce the command line arguments.

    :arg program_name: The name of the program
    :arg args: A list of the command line arguments
    :returns: A :python:`argparse.Namespace`
    :raises InvalidArgumentError: Whenever there's something wrong with the arguments.
    """
    build_parser = argparse.ArgumentParser(add_help=False)
    build_parser.add_argument(
        "ansible_version",
        type=PypiVer,
        help="The X.Y.Z version of Ansible that this will be for",
    )
    build_parser.add_argument(
        "--data-dir", default=".", help="Directory to read .build and .deps files from"
    )

    build_write_data_parser = argparse.ArgumentParser(
        add_help=False, parents=[build_parser]
    )
    build_write_data_parser.add_argument(
        "--dest-data-dir",
        default=None,
        help="Directory to write .build and .deps files to,"
        " as well as changelog and porting guide if applicable."
        "  Defaults to --data-dir",
    )
    cache_parser = argparse.ArgumentParser(add_help=False)
    cache_parser.add_argument(
        "--collection-cache",
        default=argparse.SUPPRESS,
        help="Directory of cached collection tarballs.  Will be"
        " used if a collection tarball to be downloaded exists"
        " in here, and will be populated when downloading new"
        " tarballs.",
    )

    build_step_parser = argparse.ArgumentParser(add_help=False)
    build_step_parser.add_argument(
        "--build-file",
        default=None,
        help="File containing the list of collections with version"
        " ranges.  This is considered to be relative to"
        " --build-data-dir.  The default is"
        " $DEFAULT_FILE_BASE-X.Y.build",
    )
    build_step_parser.add_argument(
        "--deps-file",
        default=None,
        help="File which will be written containing the list of"
        " collections at versions which were included in this version"
        " of Ansible.  This is considered to be relative to"
        " --build-data-dir.  The default is"
        " $BASENAME_OF_BUILD_FILE-X.Y.Z.deps",
    )
    build_step_parser.add_argument(
        "--constraints-file",
        default=None,
        help="File containing a list of constraints for collections"
        " included in Ansible.  This is considered to be relative to"
        " --build-data-dir.  The default is"
        " $BASENAME_OF_BUILD_FILE-X.Y.constraints",
    )

    feature_freeze_parser = argparse.ArgumentParser(add_help=False)
    feature_freeze_parser.add_argument(
        "--feature-frozen",
        action="store_true",
        help="If this is given, then do not allow collections whose"
        " version implies there are new features.",
    )

    galaxy_file_parser = argparse.ArgumentParser(add_help=False)
    galaxy_file_parser.add_argument(
        "--galaxy-file",
        default=None,
        help="Galaxy galaxy-requirements.yaml style file which will be"
        " written containing the list of collections at versions which"
        " were included in this version of Ansible.  This is"
        " considered to be relative to --build-data-dir.  The default"
        " is $BASENAME_OF_BUILD_FILE-X.Y.Z.yaml",
    )

    # Delay import to avoid potential import loops
    from antsibull import __version__ as _ver  # pylint: disable=import-outside-toplevel

    parser = get_toplevel_parser(
        prog=program_name,
        package="antsibull",
        description="Script to manage building Ansible",
        package_version=_ver,
    )

    subparsers = parser.add_subparsers(
        title="Subcommands",
        dest="command",
        help="for help use antsibull-build SUBCOMMANDS -h",
    )
    subparsers.required = True

    new_parser = subparsers.add_parser(
        "new-ansible",
        parents=[build_write_data_parser],
        description="Generate a new build description from the"
        " latest available versions of ansible-core and the"
        " included collections",
    )
    new_parser.add_argument(
        "--pieces-file",
        default=None,
        help="File containing a list of collections to include.  This is"
        " considered to be relative to --data-dir.  The default is"
        f" {DEFAULT_PIECES_FILE}",
    )
    new_parser.add_argument(
        "--build-file",
        default=None,
        help="File which will be written which contains the list"
        " of collections with version ranges.  This is considered to be"
        " relative to --dest-data-dir.  The default is"
        " $BASENAME_OF_PIECES_FILE-X.Y.build",
    )
    new_parser.add_argument(
        "--allow-prereleases",
        action="store_true",
        default=False,
        help="Allow prereleases of collections to be included in the build" " file",
    )
    new_parser.add_argument(
        "--constraints-file",
        default=None,
        help="File containing a list of constraints for collections"
        " included in Ansible.  This is considered to be relative to"
        " --build-data-dir.  The default is"
        " $BASENAME_OF_PIECES_FILE-X.Y.constraints",
    )

    prepare_parser = subparsers.add_parser(
        "prepare",
        parents=[
            build_write_data_parser,
            build_step_parser,
            feature_freeze_parser,
            galaxy_file_parser,
        ],
        description="Collect dependencies for an Ansible release",
    )
    prepare_parser.add_argument(
        "--tags-file",
        nargs="?",
        const="DEFAULT",
        help="Whether to include a tags data file in --dest-data-dir."
        " By default, the tags data file is stored in --dest-data-dir"
        f" as {DEFAULT_FILE_BASE}-X.Y.Z-tags.yaml."
        " --tags-file takes an optional argument to change the filename.",
    )

    build_single_parser = subparsers.add_parser(
        "single",
        parents=[
            build_write_data_parser,
            cache_parser,
            build_step_parser,
            feature_freeze_parser,
            galaxy_file_parser,
        ],
        description="Build a single-file Ansible" " [deprecated]",
    )
    build_single_parser.add_argument(
        "--sdist-dir",
        default=".",
        help="Directory to write the generated sdist tarball to",
    )
    build_single_parser.add_argument(
        "--debian",
        action="store_true",
        help="Include Debian/Ubuntu packaging files in"
        " the resulting output directory",
    )
    build_single_parser.add_argument(
        "--tags-file",
        nargs="?",
        const="DEFAULT",
        help="Whether to include a tags data file in --dest-data-dir and the sdist."
        " By default, the tags data file is stored in --dest-data-dir"
        f" as {DEFAULT_FILE_BASE}-X.Y.Z-tags.yaml."
        " --tags-file takes an optional argument to change the filename."
        " The tags data file in the sdist is always named 'tags.yaml'",
    )

    package_file_parser = argparse.ArgumentParser(add_help=False)
    package_file_parser.add_argument(
        "--debian",
        action="store_true",
        help="Include Debian/Ubuntu packaging files in"
        " the resulting output directory",
    )
    package_file_parser.add_argument(
        "--tags-file",
        nargs="?",
        const="DEFAULT",
        help="Whether to include a tags data file in the sdist."
        " By default, the tags data file is stored in --data-dir"
        f" as {DEFAULT_FILE_BASE}-X.Y.Z-tags.yaml."
        " --tags-file takes an optional argument to change the filename."
        " The tags data file in the sdist is always named 'tags.yaml'",
    )

    rebuild_single_parser = subparsers.add_parser(
        "rebuild-single",
        parents=[
            build_write_data_parser,
            cache_parser,
            build_step_parser,
            package_file_parser,
        ],
        description="Rebuild a single-file Ansible from" " a dependency file",
    )
    rebuild_single_parser.add_argument(
        "--sdist-dir",
        default=".",
        help="Directory to write the generated sdist tarball to",
    )
    rebuild_single_parser.add_argument(
        "--sdist-src-dir",
        help="Copy the files from which the source distribution is"
        " created to the specified directory. This is mainly useful"
        " for debugging antsibull-build",
    )

    subparsers.add_parser(
        "changelog",
        parents=[build_write_data_parser, cache_parser],
        description="Build the Ansible changelog",
    )

    validate_deps = subparsers.add_parser(
        "validate-deps", description="Validate collection dependencies"
    )

    validate_deps.add_argument(
        "collection_root",
        help="Path to a ansible_collections directory containing a"
        " collection tree to check.",
    )
    validate_tags_shared = argparse.ArgumentParser(add_help=False)
    validate_tags_shared.add_argument(
        "-I",
        "--ignore",
        action="append",
        help="Ignore these collections when reporting errors.",
        default=[],
    )
    validate_tags_shared.add_argument(
        "--ignores-file",
        help="Path to a file with newline separated list of collections to ignore",
        type=argparse.FileType("r"),
    )
    validate_tags_shared.add_argument(
        "-E",
        "--error-on-useless-ignores",
        action=BooleanOptionalAction,
        dest="error_on_useless_ignores",
        default=True,
        help="By default, useless ignores (e.g. passing"
        " `--ignore collection.collection` when that collection is"
        " properly tagged) will be considered an error.",
    )

    validate_tags = subparsers.add_parser(
        "validate-tags",
        parents=[build_parser, validate_tags_shared],
        description="Ensure that collection versions in an Ansible release are tagged"
        " in collections' respective git repositories.",
    )
    validate_tags.add_argument(
        "--deps-file",
        default=None,
        help="File which contains the list of collections and"
        " versions which were included in this version of Ansible."
        "  This is considered to be relative to --data-dir."
        f"  The default is {DEFAULT_FILE_BASE}-X.Y.Z.deps",
    )
    validate_tags.add_argument(
        "-o",
        "--output",
        help="Path to output a collection tag data file."
        " If this is ommited, no tag data will be written",
    )

    validate_tags_file = subparsers.add_parser(
        "validate-tags-file",
        parents=[validate_tags_shared],
        description="Ensure that collection versions in an Ansible release are tagged"
        " in collections' respective git repositories."
        " This validates the tags file generated by"
        " the 'validate-tags' subcommand.",
    )
    validate_tags_file.add_argument("tags_file")

    generate_package_files = subparsers.add_parser(
        "generate-package-files",
        description=generate_package_files_command.__doc__,
        parents=[build_parser, build_step_parser, package_file_parser],
    )
    generate_package_files.add_argument(
        "-p",
        "--package-dir",
        required=True,
        help="Directory in which to write the package files",
    )
    generate_package_files.add_argument(
        "-c", "--collections-dir", help="Defaults to {PACKAGE_DIR}/ansible_collections"
    )

    verify_upstream_parser = subparsers.add_parser(
        "verify-upstreams",
        parents=[
            cache_parser,
        ],
        description=verify_upstream_command.__doc__,
    )
    verify_upstream_parser.add_argument("tags_file")
    verify_upstream_parser.add_argument(
        "-g",
        "--glob",
        dest="globs",
        action="append",
        help="Only check collections that match the glob(s)",
    )
    _choices = [v.name for v in FileError] + [DISABLE_VERIFY_UPSTREAMS_IGNORES_SENTINEL]
    verify_upstream_parser.add_argument(
        "-I",
        "--ignore",
        action="append",
        dest="ignores",
        type=FileError,
        help=f"List of upstream verification errors to ignore. Choices: {_choices}."
        f" Default: {[v.name for v in LENIENT_FILE_ERROR_IGNORES]}.",
    )
    verify_upstream_parser.add_argument(
        "-O",
        "--error-output",
        type=Path,
        help="Path to a file to output errors",
        required=True,
    )
    verify_upstream_parser.add_argument(
        "--tree-dir",
        type=Path,
        help="Directory in which to create a collection tree",
    )
    verify_upstream_parser.add_argument(
        "--checkouts-dir",
        type=Path,
        help="Directory in which to clone collection repositories",
    )
    # Private argument to use a special download directory
    verify_upstream_parser.add_argument(
        "--download-dir", help=argparse.SUPPRESS, type=Path
    )

    sanity_test_parser = subparsers.add_parser(
        "sanity-tests",
        help=sanity_tests_command.__doc__,
    )
    sanity_test_parser.add_argument("collection_paths", type=Path, nargs="+")
    sanity_test_parser.add_argument(
        "-O",
        "--error-output",
        type=Path,
        help="Path to a YAML file to output errors",
        required=True,
    )
    sanity_test_parser.add_argument(
        "-t",
        "--test",
        action="append",
        help=f"Sanity tests to run. Default: {SANITY_TESTS_DEFAULT}",
        dest="tests",
    )
    sanity_test_parser.add_argument(
        "--clean",
        action=BooleanOptionalAction,
        default=True,
        help="Whether to clean collections' test output directories."
        " Default: %(default)s",
    )
    sanity_test_parser.add_argument(
        "--quiet",
        action=BooleanOptionalAction,
        default=False,
        help="Whether to show sanity test output as it runs. Default: %(default)s",
    )
    sanity_test_parser.add_argument("--ansible-test-bin", default="ansible-test")

    announcements_parser = subparsers.add_parser(
        "announcements",
        parents=[build_parser, build_step_parser],
        description=announcements_command.__doc__,
    )
    announcements_parser.add_argument(
        "-O", "--output-dir", type=Path, default=DEFAULT_ANNOUNCEMENTS_DIR
    )
    announcements_parser.add_argument(
        "--dist-dir",
        help="Directory containing dists to match against those uploaded to PyPI",
        type=Path,
    )
    announcements_parser.add_argument(
        "--send",
        action="store_true",
        help="Interactively send announcements using send-announcements command"
        " (with default options) after generating them",
    )

    send_announcements_parser = subparsers.add_parser(
        "send-announcements", description=send_announcements_command.__doc__
    )
    send_announcements_parser.add_argument(
        "--announcements-dir", type=Path, default=DEFAULT_ANNOUNCEMENTS_DIR
    )
    send_announcements_parser.add_argument(
        "--clipboard",
        default=True,
        action=BooleanOptionalAction,
        help="Whether to allow the command to write to the system clipboard."
        " Default: %(default)s",
    )
    send_announcements_parser.add_argument(
        "-A",
        "--action",
        action="append",
        choices=list(ALLOWED_SEND_ACTIONS),
        help="Which actions to perform."
        " --action can be specified multiple times."
        " Defaults to performing all actions.",
        dest="send_actions",
    )

    parsed_args: argparse.Namespace = parser.parse_args(args)

    # Validation and coercion
    normalize_toplevel_options(parsed_args)
    _normalize_commands(parsed_args)
    _normalize_build_options(parsed_args)
    _normalize_build_write_data_options(parsed_args)
    _normalize_new_release_options(parsed_args)
    _normalize_release_build_options(parsed_args)
    _normalize_validate_tags_options(parsed_args)
    _normalize_release_rebuild_options(parsed_args)
    _normalize_validate_tags_file_options(parsed_args)
    _normalize_generate_package_files_options(parsed_args)
    _normalize_verify_upstream_options(parsed_args)
    _normalize_announcements_options(parsed_args)
    _normalize_send_announcements_options(parsed_args)

    return parsed_args


def run(args: list[str]) -> int:
    """
    Run the program.

    :arg args: A list of command line arguments.  Typically :python:`sys.argv`.
    :returns: A program return code.  0 for success, integers for any errors.  These are documented
        in :func:`main`.
    """
    flog = mlog.fields(func="run")
    flog.fields(raw_args=args).info("Enter")

    program_name = os.path.basename(args[0])
    try:
        parsed_args: argparse.Namespace = parse_args(program_name, args[1:])
    except InvalidArgumentError as e:
        print(e)
        return 2

    try:
        cfg = load_config(parsed_args.config_file)
        flog.fields(config=cfg).info("Config loaded")
    except ConfigError as e:
        print(e)
        return 2

    context_data = app_context.create_contexts(args=parsed_args, cfg=cfg)
    with app_context.app_and_lib_context(context_data) as (app_ctx, dummy_):
        # TODO: Call `model_dump()` instead of deprecated `dict()`
        # once support for pydantic v1/antsibull-core v2 is dropped
        twiggy.dict_config(app_ctx.logging_cfg.dict())
        flog.debug("Set logging config")

        flog.fields(command=parsed_args.command).info("Action")
        return ARGS_MAP[parsed_args.command]()


def main() -> int:
    """
    Entrypoint called from the script.

    console_scripts call functions which take no parameters.  However, it's hard to test a function
    which takes no parameters so this function lightly wraps :func:`run`, which actually does the
    heavy lifting.

    :returns: A program return code.

    Return codes:
        :0: Success
        :1: Unhandled error.  See the Traceback for more information.
        :2: There was a problem with the command line arguments
        :3: version in an input file does not match with the version specified on the command line
        :4: Needs to be run on a newer version of Python
    """
    return run(sys.argv)


if __name__ == "__main__":
    sys.exit(main())



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Types used in the antsibull codebase
"""

from __future__ import annotations

import dataclasses
from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING, Any, TypeVar

import yaml
from antsibull_core.yaml import load_yaml_file

if TYPE_CHECKING:
    from _typeshed import StrPath

_T = TypeVar("_T")


class CollectionName(str):
    """
    String subclass that represents a collection with a NAMESPACE.NAME
    """

    __slots__ = ("__namespace", "__name")

    def __init__(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
        namespace, _, name = self.partition(".")
        if not namespace or not name:
            raise ValueError(f"{self!r} is not a valid collection name")
        # str should be immutable. Make these private and expose via `@property`s.
        self.__namespace = namespace
        self.__name = name

    @property
    def namespace(self) -> str:
        """
        Collection namespace
        """
        return self.__namespace

    @property
    def name(self) -> str:
        """
        Collection name
        """
        return self.__name

    @property
    def parts(self) -> tuple[str, str]:
        """
        Returns a tuple of (self.namespace, self.name)
        """
        return self.namespace, self.name

    @classmethod
    def from_pair(cls, namespace: str, name: str) -> CollectionName:
        """
        Construct a `CollectionName` object from a `namespace` and `name`
        """
        return cls(".".join((namespace, name)))

    @classmethod
    def from_galaxy_yml(cls, path: StrPath) -> CollectionName:
        """
        Construct a `CollectionName` object from a `galaxy.yml` file
        """
        data = load_yaml_file(path)
        return cls.from_pair(data["namespace"], data["name"])

    def __hash__(self) -> int:
        return hash(type(self)) + super().__hash__()


def add_yaml_type(
    typ: type[_T],
    converter: Callable[[_T], Any] = str,
    representer_parent: Callable = yaml.representer.SafeRepresenter.represent_str,
) -> None:
    """
    Add a type to the YAML serializer. Defaults to serializing as a string.
    """
    dumpers: list[type[Any]] = [yaml.SafeDumper]
    try:
        cdumper = yaml.CSafeDumper
    except AttributeError:
        pass
    else:
        dumpers.append(cdumper)

    def representer(rep: Any, obj: Any) -> Any:
        return representer_parent(rep, converter(obj))

    for dumper in dumpers:
        dumper.add_representer(typ, representer)


add_dataclass_yaml_type = partial(
    add_yaml_type,
    converter=dataclasses.asdict,
    representer_parent=lambda dumper, data: yaml.representer.SafeRepresenter.represent_mapping(
        dumper, "tag:yaml.org,2002:map", data
    ),
)


add_yaml_type(CollectionName)


def make_collection_mapping(mapping: dict[str, _T]) -> dict[CollectionName, _T]:
    """
    Convert `str` keys in a mapping to `CollectionName` objects
    """
    return {CollectionName(collection): value for collection, value in mapping.items()}


__all__ = (
    "CollectionName",
    "add_yaml_type",
    "add_dataclass_yaml_type",
    "make_collection_mapping",
)



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

# This file is here so that pytest won't complain that it cannot find tests.
# Remove once we have actual tests.

import antsibull


def test_dummy():
    pass



# Copyright (C) 2023 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Vendored or modified code from other sources
"""

from __future__ import annotations



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

import docutils.utils


def check(source: str,
          filename: str = ...,
          report_level: docutils.utils.Reporter = ...,
          ignore: dict | None = ...,
          debug: bool = ...) -> list[tuple[int, str]]: ...



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)


from __future__ import annotations

import pytest

from antsibull.types import CollectionName, make_collection_mapping


def test_collection_name() -> None:
    collection_obj = CollectionName("community.general")
    assert (
        (collection_obj.namespace, collection_obj.name)
        == ("community", "general")
        == collection_obj.parts
    )
    hash(collection_obj)


@pytest.mark.parametrize(
    "collection,error_msg",
    [
        pytest.param("abc.", "'abc.' is not a valid collection name"),
        pytest.param(".abc", "'.abc' is not a valid collection name"),
    ],
)
def test_collection_name_error(collection: str, error_msg: str) -> None:
    with pytest.raises(ValueError, match=error_msg):
        CollectionName(collection)


def test_make_collection_mapping() -> None:
    mapping = {"abc.abc": 1, "xyz.xyz": 1}
    expected = {CollectionName("abc.abc"): 1, CollectionName("xyz.xyz"): 1}
    assert make_collection_mapping(mapping) == expected



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020
"""Build command for creating a new major Ansible version."""

from __future__ import annotations

import asyncio
import os

from antsibull_core import app_context
from antsibull_core.dependency_files import BuildFile, parse_pieces_file
from packaging.version import Version as PypiVer

from .changelog import ChangelogData
from .utils.galaxy import create_galaxy_context
from .versions import (
    find_latest_compatible,
    get_version_info,
    load_constraints_if_exists,
)


def new_ansible_command() -> int:
    app_ctx = app_context.app_ctx.get()
    lib_ctx = app_context.lib_ctx.get()
    collections = parse_pieces_file(
        os.path.join(app_ctx.extra["data_dir"], app_ctx.extra["pieces_file"])
    )
    galaxy_context = asyncio.run(create_galaxy_context())
    ansible_core_release_infos, collections_to_versions = asyncio.run(
        get_version_info(
            collections, str(lib_ctx.pypi_url), galaxy_context=galaxy_context
        )
    )
    ansible_core_versions = [
        (PypiVer(version), data[0]["requires_python"])
        for version, data in ansible_core_release_infos.items()
    ]
    ansible_core_versions.sort(reverse=True, key=lambda ver_req: ver_req[0])

    constraints_filename = os.path.join(
        app_ctx.extra["data_dir"], app_ctx.extra["constraints_file"]
    )
    constraints = load_constraints_if_exists(constraints_filename)

    ansible_core_version, python_requires = ansible_core_versions[0]
    dependencies = find_latest_compatible(
        ansible_core_version,
        collections_to_versions,
        pre=app_ctx.extra["allow_prereleases"],
        prefer_pre=True,
        constraints=constraints,
    )

    build_filename = os.path.join(
        app_ctx.extra["dest_data_dir"], app_ctx.extra["build_file"]
    )
    build_file = BuildFile(build_filename)
    build_file.write(
        app_ctx.extra["ansible_version"],
        str(ansible_core_version),
        dependencies,
        python_requires=python_requires,
    )

    changelog = ChangelogData.ansible(app_ctx.extra["dest_data_dir"])
    changelog.changes.save()

    return 0



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Clone and verify collections from source
"""

from __future__ import annotations

from .commands import verify_upstream_command

__all__ = ("verify_upstream_command",)



# Copyright (C) 2023 Felix Fontein <felix@fontein.de>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

import pytest

from antsibull.build_ansible_commands import feature_freeze_version


@pytest.mark.parametrize(
    "spec, expected",
    [
        pytest.param(
            ">=1.0.0,<2.0.0",
            ">=1.0.0,<1.1.0",
            id="simple",
        ),
        pytest.param(
            "<2.0.0,>=1.0.0",
            ">=1.0.0,<1.1.0",
            id="reverted-order",
        ),
        pytest.param(
            ">=1.2.3,<1.2.7",
            ">=1.2.3,<1.2.7",
            id="tiny interval",
        ),
        pytest.param(
            ">=1.2.3,<=1.2.7",
            ">=1.2.3,<=1.2.7",
            id="tiny interval",
        ),
        pytest.param(
            ">=1.2.3,<1.3.0",
            ">=1.2.3,<1.3.0",
            id="minor half-open interval",
        ),
        pytest.param(
            ">=1.2.3,<=1.3.0",
            ">=1.2.3,<1.3.0",
            id="minor closed interval",
        ),
        pytest.param(
            "==1.2.3",
            "==1.2.3",
            id="pinned version",
        ),
        pytest.param(
            ">=1.0.0,!=1.0.1,!=1.1.2,!=1.1.0,<2.0.0",
            "!=1.0.1,>=1.0.0,<1.1.0",
            id="exclusions",
        ),
    ],
)
def test_feature_freeze_version(
    spec: str,
    expected: str,
):
    result = feature_freeze_version(spec, "foo.bar")
    assert result == expected


@pytest.mark.parametrize(
    "spec, expected",
    [
        pytest.param(
            ">=1.0.0",
            "No upper version limit specified for foo.bar: >=1.0.0",
            id="no upper limit",
        ),
        pytest.param(
            ">1.0.0",
            "Strict lower bound specified for foo.bar: >1.0.0",
            id="strict lower bound",
        ),
        pytest.param(
            ">=0.1.0,<=1.0.0,<2.0.0",
            "Multiple upper version limits specified for foo.bar: >=0.1.0,<=1.0.0,<2.0.0",
            id="multiple upper versions",
        ),
        pytest.param(
            "<1.0.0",
            "No minimum version specified for foo.bar: <1.0.0",
            id="no minimum",
        ),
        pytest.param(
            ">=1.0.0,>=2.0.0",
            "Multiple minimum versions specified for foo.bar: >=1.0.0,>=2.0.0",
            id="multiple minimum versions",
        ),
        pytest.param(
            "==1.0.0,<2.0.0",
            "Pin combined with other clauses for foo.bar: ==1.0.0,<2.0.0",
            id="pin combined with others",
        ),
    ],
)
def test_feature_freeze_version_fail(
    spec: str,
    expected: str,
):
    with pytest.raises(ValueError) as exc_info:
        result = feature_freeze_version(spec, "foo.bar")
    assert exc_info.value.args[0] == expected



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Constants used throughout the antsibull codebase
"""

from __future__ import annotations

from packaging.version import Version as PypiVer

MINIMUM_ANSIBLE_VERSION = PypiVer("6.0.0")
MINIMUM_ANSIBLE_VERSIONS = {
    "PACKAGE_DATA_NEW_METHOD": PypiVer("8.0.0a1"),
    # Whether to store setuptools config in setup.cfg
    "BUILD_META_MAKER": PypiVer("9.0.0.dev0"),
    "BUILD_META_NEW_URLS": PypiVer("9.0.0rc1"),
}

DOCSITE_BASE_URL = "https://docs.ansible.com/ansible"
DOCSITE_COMMUNITY_URL = "https://docs.ansible.com/ansible/latest/community"
BUILD_DATA_URL = "https://github.com/ansible-community/ansible-build-data"
ANSIBLE_FORUM_URL = "https://forum.ansible.com"

COLLECTION_EXCLUDE_DIRS = ("docs", "tests")

SANITY_TESTS_DEFAULT: tuple[str, ...] = (
    "ansible-doc",
    "compile",
    "validate-modules",
    "yamllint",
)
SANITY_TESTS_BANNED_IGNORES = frozenset(
    {
        "validate-modules!skip",
        "validate-modules:doc-choices-do-not-match-spec",
        "validate-modules:doc-default-does-not-match-spec",
        "validate-modules:doc-missing-type",
        "validate-modules:doc-required-mismatch",
        "validate-modules:mutually_exclusive-unknown",
        "validate-modules:no-log-needed",
        "validate-modules:nonexistent-parameter-documented",
        "validate-modules:parameter-list-no-elements",
        "validate-modules:parameter-type-not-in-doc",
        # Don't enforce this for now. Modules may have private parameters that
        # are only used by a corresponding action plugin.
        # "validate-modules:undocumented-parameter",
    }
)



# Author: Felix Fontein <felix@fontein.de>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020
"""Build Ansible changelog and porting guide."""

from __future__ import annotations

import asyncio
import os
import os.path
import typing as t
from dataclasses import dataclass

from antsibull_changelog.changelog_generator import (
    ChangelogEntry as ChangelogGeneratorEntry,
)
from antsibull_changelog.config import DEFAULT_SECTIONS, TextFormat
from antsibull_changelog.rendering.changelog import (
    add_section_content,
    create_document_renderer,
)
from antsibull_changelog.rendering.document import (
    AbstractRenderer,
    DocumentRenderer,
    SectionRenderer,
)
from antsibull_changelog.rendering.rst_document import RSTDocumentRenderer
from antsibull_core import app_context
from antsibull_core.logging import log
from packaging.version import Version as PypiVer

from .changelog import Changelog, ChangelogData, ChangelogEntry, get_changelog
from .collection_meta import CollectionsMetadata
from .utils.galaxy import create_galaxy_context

mlog = log.fields(mod=__name__)


class SectionAdder:
    parent: SectionAdder | AbstractRenderer
    section_name: str
    section: SectionRenderer | None

    def __init__(self, parent: SectionAdder | AbstractRenderer, section_name: str):
        self.parent = parent
        self.section_name = section_name
        self.section = None

    def get_section(self) -> SectionRenderer:
        if self.section is None:
            parent: AbstractRenderer
            if isinstance(self.parent, SectionAdder):
                parent = self.parent.get_section()
            else:
                parent = t.cast(AbstractRenderer, self.parent)
            self.section = parent.add_section(self.section_name)
        return self.section

    def close(self) -> None:
        if self.section is not None:
            self.section.close()


#
# Changelog
#

CHANGELOG_FORMATS = [TextFormat.RESTRUCTURED_TEXT, TextFormat.MARKDOWN]

PluginDataT = t.List[t.Tuple[str, str, t.Optional[ChangelogGeneratorEntry]]]


def _cleanup_plugins(entries: list[t.Any]) -> list[t.Any]:
    """Remove duplicate module/plugin/object entries from the given list."""
    result = []
    found_names = set()
    for entry in entries:
        name = entry["name"]
        if name in found_names:
            continue
        result.append(entry)
        found_names.add(name)
    return result


def optimize_release_entry(entry: ChangelogGeneratorEntry) -> ChangelogGeneratorEntry:
    """Remove duplicate entries from changelog entry.

    This can happen if entries from versions from different changelogs are combined
    which partially include each other's changes. This happens if for example the 1.1.0
    changelog continues the 1.0.2 changelog, and there was also a 1.0.3 release continuing
    the 1.0.2 changelog. The 1.0.3 changes are usually a subset of the 1.1.0 changes as
    they are backported bugfixes.
    """
    entry.modules = _cleanup_plugins(entry.modules)
    for plugin_type, plugins in entry.plugins.items():
        entry.plugins[plugin_type] = _cleanup_plugins(plugins)
    for object_type, objects in entry.objects.items():
        entry.objects[object_type] = _cleanup_plugins(objects)
    for section, changes in entry.changes.items():
        if isinstance(changes, list):
            entry.changes[section] = sorted(set(changes))
    return entry


def _add_table_row(
    lines: list[str], column_widths: list[int], row: list[str], sep: str
):
    row_lines = [[""] * len(column_widths)]
    for i, value in enumerate(row):
        for j, line in enumerate(value.splitlines()):
            if j >= len(row_lines):
                row_lines.append([""] * len(column_widths))
            row_lines[j][i] = line
    for cells in row_lines:
        parts = [sep]
        for j, cell in enumerate(cells):
            parts.append(" ")
            parts.append(cell)
            parts.append(" " * (1 + column_widths[j] - len(cell)))
            parts.append(sep)
        lines.append("".join(parts))


def _add_table_line(
    lines: list[str],
    column_widths: list[int],
    start_sep: str,
    mid_sep: str,
    end_sep: str,
    corner: str,
):
    parts = [corner]
    for w in column_widths:
        parts.append(start_sep)
        parts.append(mid_sep * w)
        parts.append(end_sep)
        parts.append(corner)
    lines.append("".join(parts))


def compute_column_widths(headings: list[str], cells: list[list[str]]) -> list[int]:
    column_widths: list[int] = []
    for row in [headings] + cells:
        while len(row) > len(column_widths):
            column_widths.append(0)
        for i, value in enumerate(row):
            for line in value.splitlines():
                column_widths[i] = max(column_widths[i], len(line))
    return column_widths


def render_rst_table(headings: list[str], cells: list[list[str]]) -> str:
    column_widths = compute_column_widths(headings, cells)
    lines: list[str] = []
    _add_table_line(lines, column_widths, "-", "-", "-", "+")
    _add_table_row(lines, column_widths, headings, "|")
    _add_table_line(lines, column_widths, "=", "=", "=", "+")
    for row in cells:
        _add_table_row(lines, column_widths, row, "|")
        _add_table_line(lines, column_widths, "-", "-", "-", "+")
    return "\n".join(lines)


def render_md_table(headings: list[str], cells: list[list[str]]) -> str:
    column_widths = compute_column_widths(headings, cells)
    lines: list[str] = []
    _add_table_row(lines, column_widths, headings, "|")
    _add_table_line(lines, column_widths, " ", "-", " ", "|")
    for row in cells:
        _add_table_row(lines, column_widths, row, "|")
    return "\n".join(lines)


def format_link(title: str, url: str, text_format: TextFormat) -> str:
    if text_format == TextFormat.RESTRUCTURED_TEXT:
        return f"`{title} <{url}>`_"
    if text_format == TextFormat.MARKDOWN:
        return f"`[{title}]({url})"
    raise ValueError(f"Unknown format {format}")


def render_table(
    renderer: AbstractRenderer,
    headings: list[str],
    cells: list[list[str]],
    text_format: TextFormat,
) -> None:
    table_renderers = {
        TextFormat.RESTRUCTURED_TEXT: render_rst_table,
        TextFormat.MARKDOWN: render_md_table,
    }
    table_renderer = table_renderers.get(text_format)
    if table_renderer is None:
        raise ValueError(f"Unknown format {text_format}")
    renderer.add_text(table_renderer(headings, cells), text_format=text_format)


def append_changelog_changes_collections(
    renderer: AbstractRenderer,
    collection_metadata: CollectionsMetadata,
    changelog_entry: ChangelogEntry,
    is_last: bool,
    text_format: TextFormat,
) -> PluginDataT:
    result: PluginDataT = []

    if changelog_entry.changed_collections:
        section = renderer.add_section(
            "Included Collections" if is_last else "Changed Collections"
        )
        section.add_text(
            "If not mentioned explicitly, the changes are reported in the combined changelog"
            " below.",
            text_format=TextFormat.RESTRUCTURED_TEXT,
        )
        section.ensure_paragraph_break()
        headings = [
            "Collection",
            f"Ansible {changelog_entry.prev_version}",
            f"Ansible {changelog_entry.version_str}",
            "Notes",
        ]
        cells = []
        for (
            collector,
            collection_version,
            prev_collection_version,
            newly_added,
        ) in changelog_entry.changed_collections:
            row = [collector.collection, "", str(collection_version), ""]
            if prev_collection_version is not None:
                row[1] = str(prev_collection_version)
            changelog = collector.changelog
            if newly_added:
                row[-1] = "The collection was added to Ansible"
            elif changelog:
                release_entries = changelog.generator.collect(
                    squash=True,
                    after_version=prev_collection_version,
                    until_version=collection_version,
                )
                if not release_entries:
                    row[-1] = "The collection did not have a changelog in this version."
                elif release_entries[0].empty:
                    row[-1] = "There are no changes recorded in the changelog."
                else:
                    result.append(
                        (
                            collector.collection,
                            f"{collector.collection}.",
                            optimize_release_entry(release_entries[0]),
                        )
                    )
            else:
                metadata = collection_metadata.get_meta(collector.collection)
                if metadata.changelog_url is not None:
                    link = format_link(
                        metadata.changelog_url,
                        metadata.changelog_url,
                        text_format=text_format,
                    )
                    row[-1] = f"You can find the collection's changelog at {link}."
                else:
                    row[-1] = (
                        "Unfortunately, this collection does not provide changelog data in a"
                        " format that can be processed by the changelog generator."
                    )
            cells.append(row)
        render_table(section, headings, cells, text_format)
        section.close()

    return result


def append_changelog_changes_ansible(
    renderer: AbstractRenderer, changelog_entry: ChangelogEntry
) -> PluginDataT:
    changelog = changelog_entry.ansible_changelog

    release_entries = changelog.generator.collect(
        squash=True,
        after_version=(
            str(changelog_entry.prev_version) if changelog_entry.prev_version else None
        ),
        until_version=changelog_entry.version_str,
    )

    if not release_entries:
        return []

    release_entry = optimize_release_entry(release_entries[0])

    release_summary = release_entry.changes.pop("release_summary", None)
    if release_summary:
        section = renderer.add_section("Release Summary")
        section.add_text(
            t.cast(str, release_summary), text_format=TextFormat.RESTRUCTURED_TEXT
        )
        section.close()

    if release_entry.empty:
        return []

    return [("", "", release_entry)]


def append_changelog_changes_core(
    renderer: AbstractRenderer, changelog_entry: ChangelogEntry
) -> PluginDataT:
    section = renderer.add_section("Ansible-core")
    try:
        section.add_text(
            f"Ansible {changelog_entry.version} contains ansible-core "
            f"version {changelog_entry.ansible_core_version}.",
            text_format=TextFormat.RESTRUCTURED_TEXT,
        )
        if changelog_entry.prev_ansible_core_version:
            if (
                changelog_entry.prev_ansible_core_version
                == changelog_entry.ansible_core_version
            ):
                section.add_text(
                    "This is the same version of ansible-core as in "
                    "the previous Ansible release.",
                    text_format=TextFormat.RESTRUCTURED_TEXT,
                )
                return []

            section.add_text(
                f"This is a newer version than version "
                f"{changelog_entry.prev_ansible_core_version} contained in the "
                f"previous Ansible release.",
                text_format=TextFormat.RESTRUCTURED_TEXT,
            )

        changelog = changelog_entry.core_collector.changelog
        if not changelog:
            return []

        section.ensure_paragraph_break()

        release_entries = changelog.generator.collect(
            squash=True,
            after_version=changelog_entry.prev_ansible_core_version,
            until_version=changelog_entry.ansible_core_version,
        )

        if not release_entries:
            section.add_text(
                "Ansible-core did not have a changelog in this version.",
                text_format=TextFormat.RESTRUCTURED_TEXT,
            )
            return []

        release_entry = release_entries[0]

        if release_entry.empty:
            section.add_text(
                "There are no changes recorded in the changelog.",
                text_format=TextFormat.RESTRUCTURED_TEXT,
            )
            return []

        section.add_text(
            "The changes are reported in the combined changelog below.",
            text_format=TextFormat.RESTRUCTURED_TEXT,
        )

        return [
            (
                "Ansible-core",
                "ansible.builtin.",
                release_entry,
            )
        ]
    finally:
        section.close()


def common_start(a: list[t.Any], b: list[t.Any]) -> int:
    """
    Given two sequences a and b, determines maximal index so that
    all elements up to that index are equal.
    """
    common_len = min(len(a), len(b))
    for i in range(common_len):
        if a[i] != b[i]:
            return i
    return common_len


PluginDumpT = t.List[t.Tuple[t.List[str], str, str]]


def _get_last_renderer(
    sections: list[SectionRenderer], renderer: AbstractRenderer
) -> AbstractRenderer:
    if sections:
        return sections[-1]
    return renderer


def dump_items(renderer: AbstractRenderer, items: PluginDumpT) -> None:
    last_title: list[str] = []
    sections: list[SectionRenderer] = []
    for title, name, description in sorted(items):
        if title != last_title:
            common = common_start(last_title, title)
            while len(sections) > common:
                sections.pop().close()
            while len(sections) < len(title):
                section = _get_last_renderer(sections, renderer).add_section(
                    title[len(sections)]
                )
                sections.append(section)
            last_title = title
        _get_last_renderer(sections, renderer).add_fragment(
            f"{name} - {description}", text_format=TextFormat.RESTRUCTURED_TEXT
        )

    while sections:
        sections.pop().close()


def add_plugins(renderer: AbstractRenderer, data: PluginDataT) -> None:
    plugins: PluginDumpT = []
    for _, prefix, release_entry in data:
        if release_entry:
            for plugin_type, plugin_datas in release_entry.plugins.items():
                for plugin_data in plugin_datas:
                    plugins.append(
                        (
                            # ['New Plugins', plugin_type.title(), name],
                            ["New Plugins", plugin_type.title()],
                            prefix + plugin_data["name"],
                            plugin_data["description"],
                        )
                    )
    dump_items(renderer, plugins)


def add_objects(renderer: AbstractRenderer, data: PluginDataT) -> None:
    objects: PluginDumpT = []
    for _, prefix, release_entry in data:
        if release_entry:
            for object_type, object_datas in release_entry.objects.items():
                for object_data in object_datas:
                    objects.append(
                        (
                            [f"New {object_type.title()}s"],
                            prefix + object_data["name"],
                            object_data["description"],
                        )
                    )
    dump_items(renderer, objects)


def add_modules(renderer: AbstractRenderer, data: PluginDataT) -> None:
    modules: PluginDumpT = []
    for name, prefix, release_entry in data:
        if release_entry:
            for module in release_entry.modules:
                namespace = module.get("namespace") or ""
                if namespace.startswith(".ansible.collections.ansible_collections."):
                    # Work around old antsibull-changelog versions which suffer from
                    # https://github.com/ansible-community/antsibull-changelog/issues/18
                    namespace = ""
                namespace = namespace.strip(".").split(".", 1) if namespace else []
                modules.append(
                    (
                        ["New Modules", name]
                        + [ns.replace("_", " ").title() for ns in namespace],
                        prefix + module["name"],
                        module["description"],
                    )
                )
    dump_items(renderer, modules)


def append_removed_collections(
    renderer: AbstractRenderer, changelog_entry: ChangelogEntry
) -> None:
    if changelog_entry.removed_collections:
        section = renderer.add_section("Removed Collections")
        for collector, collection_version in changelog_entry.removed_collections:
            section.add_fragment(
                f"{collector.collection} "
                f"(previously included version: {collection_version})",
                text_format=TextFormat.RESTRUCTURED_TEXT,
            )
        section.close()


def append_added_collections(
    renderer: AbstractRenderer, changelog_entry: ChangelogEntry
) -> None:
    if changelog_entry.added_collections:
        section = renderer.add_section("Added Collections")
        for collector, collection_version in changelog_entry.added_collections:
            section.add_fragment(
                f"{collector.collection} (version {collection_version})",
                text_format=TextFormat.RESTRUCTURED_TEXT,
            )
        section.close()


def append_unchanged_collections(
    renderer: AbstractRenderer, changelog_entry: ChangelogEntry
) -> None:
    if changelog_entry.unchanged_collections:
        section = renderer.add_section("Unchanged Collections")
        for collector, collection_version in changelog_entry.unchanged_collections:
            section.add_fragment(
                f"{collector.collection} (still version {collection_version})",
                text_format=TextFormat.RESTRUCTURED_TEXT,
            )
        section.close()


def append_changelog(
    renderer: AbstractRenderer,
    collection_metadata: CollectionsMetadata,
    changelog_entry: ChangelogEntry,
    is_last: bool,
    text_format: TextFormat,
) -> None:
    section = renderer.add_section(f"v{changelog_entry.version_str}")
    section.add_toc(max_depth=2)

    # Add release summary for Ansible
    data = append_changelog_changes_ansible(section, changelog_entry)

    append_removed_collections(section, changelog_entry)
    append_added_collections(section, changelog_entry)

    # Adds Ansible-core section
    data.extend(append_changelog_changes_core(section, changelog_entry))

    # Adds list of changed collections
    data.extend(
        append_changelog_changes_collections(
            section,
            collection_metadata,
            changelog_entry,
            is_last=is_last,
            text_format=text_format,
        )
    )

    # Adds all changes
    for section_name, section_title in DEFAULT_SECTIONS:
        section_renderer = SectionAdder(section, section_title)
        for name, _, release_entry in data:
            if not release_entry or release_entry.has_no_changes([section_name]):
                continue

            subsection_renderer = None
            if name:
                subsection_renderer = section_renderer.get_section().add_section(name)
            add_section_content(
                release_entry,
                subsection_renderer or section_renderer.get_section(),
                section_name,
            )
            if subsection_renderer:
                subsection_renderer.close()

        section_renderer.close()

    # Adds new plugins and modules
    add_plugins(section, data)
    add_modules(section, data)
    add_objects(section, data)

    # Adds list of unchanged collections
    append_unchanged_collections(section, changelog_entry)

    section.close()


def _compose_changelog(changelog: Changelog, text_format: TextFormat) -> str:
    flog = mlog.fields(func="_compose_changelog")

    renderer = create_document_renderer(text_format)
    version = f"{changelog.ansible_version.major}"
    renderer.set_title(f"Ansible {version} Release Notes")

    if changelog.ansible_ancestor_version:
        renderer.add_text(
            f"This changelog describes changes since"
            f" Ansible {changelog.ansible_ancestor_version}.",
            text_format=TextFormat.RESTRUCTURED_TEXT,
        )

    renderer.add_toc(max_depth=2)

    entries = [entry for entry in changelog.entries if not entry.is_ancestor]
    for index, changelog_entry in enumerate(entries):
        append_changelog(
            renderer,
            changelog.collection_metadata,
            changelog_entry,
            is_last=index + 1 == len(entries),
            text_format=text_format,
        )

    text = renderer.render()
    for warning in renderer.get_warnings():
        flog.warning(warning)
    return text


#
# Porting Guide
#


def append_porting_guide_section(
    changelog_entry: ChangelogEntry,
    maybe_add_title: SectionAdder,
    section: str,
) -> None:
    maybe_add_section_title = SectionAdder(
        maybe_add_title, section.replace("_", " ").title()
    )

    def check_changelog(
        name: str,
        changelog: ChangelogData | None,
        version: str,
        prev_version: str | None,
    ) -> None:
        if not changelog:
            return
        entries = changelog.generator.collect(
            squash=True, after_version=prev_version, until_version=version
        )
        if not entries or entries[0].has_no_changes([section]):
            return
        subsection = maybe_add_section_title.get_section()
        subsubsection = None
        if name:
            subsubsection = subsection.add_section(name)
        add_section_content(
            optimize_release_entry(entries[0]), subsubsection or subsection, section
        )
        if subsubsection:
            subsubsection.close()

    check_changelog(
        "",
        changelog_entry.ansible_changelog,
        changelog_entry.version_str,
        str(changelog_entry.prev_version) if changelog_entry.prev_version else None,
    )
    check_changelog(
        "Ansible-core",
        changelog_entry.core_collector.changelog,
        changelog_entry.ansible_core_version,
        changelog_entry.prev_ansible_core_version,
    )
    for (
        collector,
        collection_version,
        prev_collection_version,
        newly_added,
    ) in changelog_entry.changed_collections:
        if newly_added:
            continue
        check_changelog(
            collector.collection,
            collector.changelog,
            collection_version,
            prev_collection_version,
        )

    maybe_add_section_title.close()


def append_porting_guide(
    document: DocumentRenderer, changelog_entry: ChangelogEntry
) -> None:
    maybe_add_title = SectionAdder(
        document, f"Porting Guide for v{changelog_entry.version_str}"
    )

    if changelog_entry.added_collections:
        append_added_collections(maybe_add_title.get_section(), changelog_entry)

    for section_name in ["known_issues", "breaking_changes", "major_changes"]:
        append_porting_guide_section(changelog_entry, maybe_add_title, section_name)

    if changelog_entry.removed_collections:
        section = maybe_add_title.get_section().add_section("Removed Collections")
        for collector, collection_version in changelog_entry.removed_collections:
            section.add_fragment(
                f"{collector.collection} "
                f"(previously included version: {collection_version})",
                text_format=TextFormat.RESTRUCTURED_TEXT,
            )
        section.close()

    for section_name in ["removed_features", "deprecated_features"]:
        append_porting_guide_section(changelog_entry, maybe_add_title, section_name)

    maybe_add_title.close()


#
# Release Notes
#


@dataclass
class FileWithContent:
    filename: str
    content: bytes

    def write_to(self, dest_dir: str) -> None:
        path = os.path.join(dest_dir, self.filename)
        with open(path, "wb") as fd:
            fd.write(self.content)


class ReleaseNotes:
    changelogs: list[FileWithContent]
    porting_guide: FileWithContent

    def __init__(
        self,
        changelogs: list[FileWithContent],
        porting_guide: FileWithContent,
    ):
        self.changelogs = changelogs

        self.porting_guide = porting_guide

    @staticmethod
    def _get_changelog_bytes(changelog: Changelog, text_format: TextFormat) -> bytes:
        text = _compose_changelog(changelog, text_format)
        return text.encode("utf-8")

    @staticmethod
    def _append_core_porting_guide_bytes(
        renderer: AbstractRenderer, changelog: Changelog
    ) -> None:
        core_porting_guide = changelog.core_collector.porting_guide
        if core_porting_guide:
            lines = core_porting_guide.decode("utf-8").splitlines()
            lines.append("")
            found_topics = False
            found_empty = False
            append_lines = []
            for line in lines:
                if not found_topics:
                    if line.startswith(".. contents::"):
                        found_topics = True
                    continue
                if not found_empty:
                    if line == "":
                        found_empty = True
                    continue
                append_lines.append(line)
            if not found_empty:
                print("WARNING: cannot find TOC of ansible-core porting guide!")
            if append_lines:
                renderer.add_text(
                    "\n".join(append_lines), text_format=TextFormat.RESTRUCTURED_TEXT
                )

    @staticmethod
    def _get_porting_guide_bytes(changelog: Changelog) -> bytes:
        flog = mlog.fields(func="ReleaseNotes._get_porting_guide_bytes")

        version = f"{changelog.ansible_version.major}"
        core_version_obj = changelog.core_collector.latest
        core_version = f"{core_version_obj.major}.{core_version_obj.minor}"

        document = RSTDocumentRenderer()
        document.set_raw_preamble(
            "..\n"
            f"   THIS DOCUMENT IS AUTOMATICALLY GENERATED BY ANTSIBULL! PLEASE DO NOT EDIT"
            f" MANUALLY! (YOU PROBABLY WANT TO EDIT porting_guide_core_{core_version}.rst)"
            "\n"
        )
        document.set_document_label(f"porting_{version}_guide")
        document.set_title(f"Ansible {version} Porting Guide")
        document.add_toc(max_depth=2)

        # Determine ansible-core version in previous major release
        prev_core_version = ""
        if any(entry.is_ancestor for entry in changelog.entries):
            # If there is an ancestor, the earliest ansible-core version will be the
            # version used in the previous major release.
            prev_core_version_obj = changelog.core_collector.earliest
            prev_core_version = (
                f"{prev_core_version_obj.major}.{prev_core_version_obj.minor}"
            )

        # Determine whether to include ansible-core porting guide or not
        if core_version != prev_core_version:
            document.add_text(
                # noqa: E501
                "\n" f"Ansible {version} is based on Ansible-core {core_version}." "\n",
                text_format=TextFormat.RESTRUCTURED_TEXT,
            )
            document.add_text(
                # noqa: E501
                "\n"
                f"We suggest you read this page along with the `Ansible {version} Changelog"
                f" <https://github.com/ansible-community/ansible-build-data/blob/main/{version}/"
                f"CHANGELOG-v{version}.md>`_ to understand what updates you may need to make."
                "\n",
                text_format=TextFormat.RESTRUCTURED_TEXT,
            )
            ReleaseNotes._append_core_porting_guide_bytes(document, changelog)
        else:
            # Generic message if we again have two consecutive versions with the same ansible-core
            prev_version = changelog.ansible_version.major - 1
            prev_prev_version = changelog.ansible_version.major - 2
            document.add_text(
                # noqa: E501
                "\n"
                f"Ansible {version} is based on Ansible-core {core_version}, which is the same"
                f" major release as Ansible {prev_version}.  Therefore, there is no section on"
                " ansible-core in this porting guide.  If you are upgrading from Ansible"
                f" {prev_prev_version}, please first consult the Ansible {prev_version} porting"
                f" guide before continuing with the Ansible {version} porting guide."
                "\n\n"
                f"We suggest you read this page along with the `Ansible {version} Changelog"
                f" <https://github.com/ansible-community/ansible-build-data/blob/main/{version}/"
                f"CHANGELOG-v{version}.md>`_ to understand what updates you may need to make."
                "\n",
                text_format=TextFormat.RESTRUCTURED_TEXT,
            )

        for porting_guide_entry in changelog.entries:
            if not porting_guide_entry.is_ancestor:
                append_porting_guide(document, porting_guide_entry)

        text = document.render()
        for warning in document.get_warnings():
            flog.warning(warning)

        return text.encode("utf-8")

    @classmethod
    def build(cls, changelog: Changelog) -> "ReleaseNotes":
        version = f"{changelog.ansible_version.major}"
        changelogs: list[FileWithContent] = []
        for text_format in CHANGELOG_FORMATS:
            changelogs.append(
                FileWithContent(
                    f"CHANGELOG-v{version}.{text_format.to_extension()}",
                    cls._get_changelog_bytes(changelog, text_format),
                )
            )
        return cls(
            changelogs,
            FileWithContent(
                f"porting_guide_{version}.rst", cls._get_porting_guide_bytes(changelog)
            ),
        )

    def write_changelog_to(self, dest_dir: str) -> None:
        for changelog in self.changelogs:
            changelog.write_to(dest_dir)

    def write_porting_guide_to(self, dest_dir: str) -> None:
        self.porting_guide.write_to(dest_dir)


def build_changelog() -> int:
    """Create changelog and porting guide CLI command."""
    app_ctx = app_context.app_ctx.get()
    lib_ctx = app_context.lib_ctx.get()

    ansible_version: PypiVer = app_ctx.extra["ansible_version"]
    data_dir: str = app_ctx.extra["data_dir"]
    dest_data_dir: str = app_ctx.extra["dest_data_dir"]
    collection_cache: str | None = lib_ctx.collection_cache

    galaxy_context = asyncio.run(create_galaxy_context())
    changelog = get_changelog(
        ansible_version,
        galaxy_context=galaxy_context,
        deps_dir=data_dir,
        collection_cache=collection_cache,
    )

    release_notes = ReleaseNotes.build(changelog)
    release_notes.write_changelog_to(dest_data_dir)
    release_notes.write_porting_guide_to(dest_data_dir)

    missing_changelogs = []
    last_entry = changelog.entries[0]
    last_version_collectors = [
        collector
        for collector in changelog.collection_collectors
        if last_entry.version
        in last_entry.versions_per_collection[collector.collection]
    ]
    for collector in last_version_collectors:
        if collector.changelog is None:
            missing_changelogs.append(collector.collection)
    if missing_changelogs:
        print(
            f"{len(missing_changelogs)} out of {len(last_version_collectors)} collections"
            f" have no compatible changelog:"
        )
        for collection_name in missing_changelogs:
            meta = changelog.collection_metadata.get_meta(collection_name)
            entry = [collection_name]
            if meta.changelog_url:
                entry.append(f"(changelog URL: {meta.changelog_url})")
            print(f"    {'  '.join(entry)}")
    return 0
